##########################################################################
# Copyright 2013-2021 Aerospike, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##########################################################################
"""
Map expressions contain expressions for reading and modifying Maps. Most of
these operations are from the standard :mod:`Map API <aerospike_helpers.operations.map_operations>`.

"""

# from __future__ import annotations
from typing import List, Union, Dict, Any
import aerospike
from aerospike_helpers import cdt_ctx
from aerospike_helpers.expressions.resources import _GenericExpr
from aerospike_helpers.expressions.resources import _BaseExpr
from aerospike_helpers.expressions.resources import _ExprOp
from aerospike_helpers.expressions.resources import _Keys, ReturnType
from aerospike_helpers.expressions.base import MapBin

########################
# Map Modify Expressions
########################

TypeKey = Union[_BaseExpr, Any]
TypeKeyList = Union[_BaseExpr, List[Any]]
TypeBinName = Union[_BaseExpr, str]
TypeListValue = Union[_BaseExpr, List[Any]]
TypeIndex = Union[_BaseExpr, int, aerospike.CDTInfinite]
TypeCTX = Union[None, List[cdt_ctx._cdt_ctx]]
TypeRank = Union[_BaseExpr, int, aerospike.CDTInfinite]
TypeCount = Union[_BaseExpr, int, aerospike.CDTInfinite]
TypeValue = Union[_BaseExpr, Any]
TypePolicy = Union[Dict[str, Any], None]


class MapPut(_BaseExpr):
    """Create an expression that writes key/val to map bin."""

    _op = aerospike.OP_MAP_PUT

    def __init__(self, ctx: "TypeCTX", policy: "TypePolicy", key: "TypeKey", value: "TypeValue", bin: "TypeBinName"):
        """Args:
            ctx (TypeCTX): An optional list of nested CDT :mod:`cdt_ctx <aerospike_helpers.cdt_ctx>` context operation
                objects.
            policy (TypePolicy): Optional dictionary of :ref:`Map policies <aerospike_map_policies>`.
            key (TypeKey): Key value or value expression to put into map.
            value (TypeValue): Value or value expression to put into map.
            bin (TypeBinName): bin expression, such as :class:`~aerospike_helpers.expressions.base.MapBin` or
                :class:`~aerospike_helpers.expressions.base.ListBin`.

        :return: Map expression.

        Example::

            # Put {"key": 27} into map bin "b".
            expr = exp.MapPut(None, None, "key", 27, exp.MapBin("b")).compile()
        """
        self._children = (
            key,
            value,
            _GenericExpr(
                _ExprOp._AS_EXP_CODE_CDT_MAP_CRMOD, 0, {_Keys.MAP_POLICY_KEY: policy} if policy is not None else {}
            ),
            bin if isinstance(bin, _BaseExpr) else MapBin(bin),
        )
        self._fixed = {}

        if ctx is not None:
            self._fixed[_Keys.CTX_KEY] = ctx

        if policy is not None:
            self._fixed[_Keys.MAP_POLICY_KEY] = policy


class MapPutItems(_BaseExpr):
    """Create an expression that writes each map item to map bin."""

    _op = aerospike.OP_MAP_PUT_ITEMS

    def __init__(self, ctx: "TypeCTX", policy: "TypePolicy", map: map, bin: "TypeBinName"):
        """Args:
            ctx (TypeCTX): An optional list of nested CDT :mod:`cdt_ctx <aerospike_helpers.cdt_ctx>` context operation
                objects.
            policy (TypePolicy): Optional dictionary of :ref:`Map policies <aerospike_map_policies>`.
            map (map): Map or map expression of items to put into target map.
            bin (TypeBinName): bin expression, such as :class:`~aerospike_helpers.expressions.base.MapBin` or
                :class:`~aerospike_helpers.expressions.base.ListBin`.

        :return: Map expression.

        Example::

            # Put {27: 'key27', 28: 'key28'} into map bin "b".
            expr = exp.MapPutItems(None, None, {27: 'key27', 28: 'key28'}, exp.MapBin("b")).compile()
        """
        self._children = (
            map,
            _GenericExpr(
                _ExprOp._AS_EXP_CODE_CDT_MAP_CRMOD, 0, {_Keys.MAP_POLICY_KEY: policy} if policy is not None else {}
            ),
            bin if isinstance(bin, _BaseExpr) else MapBin(bin),
        )
        self._fixed = {}

        if ctx is not None:
            self._fixed[_Keys.CTX_KEY] = ctx

        if policy is not None:
            self._fixed[_Keys.MAP_POLICY_KEY] = policy


class MapIncrement(_BaseExpr):
    """Create an expression that increments a map value, by value, for all items identified by key.
    Valid only for numbers.
    """

    _op = aerospike.OP_MAP_INCREMENT

    def __init__(self, ctx: "TypeCTX", policy: "TypePolicy", key: "TypeKey", value: "TypeValue", bin: "TypeBinName"):
        """Args:
            ctx (TypeCTX): An optional list of nested CDT :mod:`cdt_ctx <aerospike_helpers.cdt_ctx>` context operation
                objects.
            policy (TypePolicy): Optional dictionary of :ref:`Map policies <aerospike_map_policies>`.
            key (TypeKey): Key value or value expression element to increment.
            value (TypeValue): Increment element by value expression.
            bin (TypeBinName): bin expression, such as :class:`~aerospike_helpers.expressions.base.MapBin` or
                :class:`~aerospike_helpers.expressions.base.ListBin`.

        :return: Map expression.

        Example::

            # Increment element at 'vageta' in map bin "b" by 9000.
            expr = exp.MapIncrement(None, None, 'vageta', 9000, exp.MapBin("b")).compile()
        """
        self._children = (
            key,
            value,
            _GenericExpr(
                _ExprOp._AS_EXP_CODE_CDT_MAP_CRMOD, 0, {_Keys.MAP_POLICY_KEY: policy} if policy is not None else {}
            ),
            bin if isinstance(bin, _BaseExpr) else MapBin(bin),
        )
        self._fixed = {}

        if ctx is not None:
            self._fixed[_Keys.CTX_KEY] = ctx

        if policy is not None:
            self._fixed[_Keys.MAP_POLICY_KEY] = policy


class MapClear(_BaseExpr):
    """Create an expression that removes all items in map."""

    _op = aerospike.OP_MAP_CLEAR

    def __init__(self, ctx: "TypeCTX", bin: "TypeBinName"):
        """Args:
            ctx (TypeCTX): An optional list of nested CDT :mod:`cdt_ctx <aerospike_helpers.cdt_ctx>` context operation
                objects.
            bin (TypeBinName): bin expression, such as :class:`~aerospike_helpers.expressions.base.MapBin` or
                :class:`~aerospike_helpers.expressions.base.ListBin`.

        :return: Map expression.

        Example::

            # Clear map bin "b".
            expr = exp.MapClear(None, exp.MapBin("b")).compile()
        """
        self._children = (bin if isinstance(bin, _BaseExpr) else MapBin(bin),)
        self._fixed = {}

        if ctx is not None:
            self._fixed[_Keys.CTX_KEY] = ctx


class MapRemoveByKey(_BaseExpr):
    """Create an expression that removes a map item identified by key."""

    _op = aerospike.OP_MAP_REMOVE_BY_KEY

    def __init__(self, ctx: "TypeCTX", key: "TypeKey", bin: "TypeBinName"):
        """Args:
            ctx (TypeCTX): An optional list of nested CDT :mod:`cdt_ctx <aerospike_helpers.cdt_ctx>` context operation
                objects.
            key (TypeKey): Key value or value expression of key to element to remove.
            bin (TypeBinName): bin expression, such as :class:`~aerospike_helpers.expressions.base.MapBin` or
                :class:`~aerospike_helpers.expressions.base.ListBin`.

        :return: Map expression.

        Example::

            # Remove element at key 1 in map bin "b".
            expr = exp.MapRemoveByKey(None, 1, exp.MapBin("b")).compile()
        """
        self._children = (
            key,
            bin if isinstance(bin, _BaseExpr) else MapBin(bin),
        )
        self._fixed = {}

        if ctx is not None:
            self._fixed[_Keys.CTX_KEY] = ctx


class MapRemoveByKeyList(_BaseExpr):
    """Create an expression that removes map items identified by keys."""

    _op = aerospike.OP_MAP_REMOVE_BY_KEY_LIST

    def __init__(self, ctx: "TypeCTX", keys: List[TypeKey], bin: "TypeBinName", inverted: bool = False):
        """Args:
            ctx (TypeCTX): An optional list of nested CDT :mod:`cdt_ctx <aerospike_helpers.cdt_ctx>` context operation
                objects.
            key (List[TypeKey]): List of key values or a list expression of keys to elements to remove.
            bin (TypeBinName): bin expression, such as :class:`~aerospike_helpers.expressions.base.MapBin` or
                :class:`~aerospike_helpers.expressions.base.ListBin`.
            inverted (bool): Invert the expression's search criteria.

        :return: Map expression.

        Example::

            # Remove elements at keys [1, 2] in map bin "b".
            expr = exp.MapRemoveByKeyList(None, [1, 2], exp.MapBin("b")).compile()
        """
        self._children = (
            keys,
            bin if isinstance(bin, _BaseExpr) else MapBin(bin),
        )
        self._fixed = {_Keys.RETURN_TYPE_KEY: ReturnType.MAP_RETURN_INVERTED if inverted else aerospike.MAP_RETURN_NONE}

        if ctx is not None:
            self._fixed[_Keys.CTX_KEY] = ctx


class MapRemoveByKeyRange(_BaseExpr):
    """Create an expression that removes map items identified by key range
    (begin inclusive, end exclusive). If begin is None, the range is less than end.
    If end is None, the range is greater than equal to begin.
    """

    _op = aerospike.OP_MAP_REMOVE_BY_KEY_RANGE

    def __init__(
        self,
        ctx: "TypeCTX",
        begin: "TypeValue",
        end: "TypeValue",
        bin: "TypeBinName",
        inverted: bool = False
    ):
        """Args:
            ctx (TypeCTX): An optional list of nested CDT :mod:`cdt_ctx <aerospike_helpers.cdt_ctx>` context operation
                objects.
            begin (TypeValue): Begin value expression.
            end (TypeValue): End value expression.
            bin (TypeBinName): bin expression, such as :class:`~aerospike_helpers.expressions.base.MapBin` or
                :class:`~aerospike_helpers.expressions.base.ListBin`.
            inverted (bool): Invert the expression's search criteria.

        :return: Map expression.

        Example::

            # Remove elements at keys between 1 and 10 in map bin "b".
            expr = exp.MapRemoveByKeyRange(None, 1, 10, exp.MapBin("b")).compile()
        """
        self._children = (
            begin,
            end,
            bin if isinstance(bin, _BaseExpr) else MapBin(bin),
        )
        self._fixed = {_Keys.RETURN_TYPE_KEY: ReturnType.MAP_RETURN_INVERTED if inverted else aerospike.MAP_RETURN_NONE}

        if ctx is not None:
            self._fixed[_Keys.CTX_KEY] = ctx


class MapRemoveByKeyRelIndexRangeToEnd(_BaseExpr):
    """Create an expression that removes map items nearest to key and greater by index."""

    _op = aerospike.OP_MAP_REMOVE_BY_KEY_REL_INDEX_RANGE_TO_END

    def __init__(self, ctx: "TypeCTX", key: "TypeKey", index: "TypeIndex", bin: "TypeBinName", inverted: bool = False):
        """Args:
            ctx (TypeCTX): An optional list of nested CDT :mod:`cdt_ctx <aerospike_helpers.cdt_ctx>` context operation
                objects.
            key (TypeKey): Key value or expression for key to start removing from.
            index (TypeIndex): Index integer or integer expression.
            bin (TypeBinName): bin expression, such as :class:`~aerospike_helpers.expressions.base.MapBin` or
                :class:`~aerospike_helpers.expressions.base.ListBin`.
            inverted (bool): Invert the expression's search criteria.

        :return: Map expression.

        Example::

            # {"key1": 1, "key2": 2, "key3": 3, "key4": 10}
            expr = exp.MapRemoveByKeyRelIndexRangeToEnd(None, "key2", 1, exp.MapBin("b")).compile()
            # This returns {"key1": 1, "key2": 2}
        """
        self._children = (
            key,
            index,
            bin if isinstance(bin, _BaseExpr) else MapBin(bin),
        )
        self._fixed = {_Keys.RETURN_TYPE_KEY: ReturnType.MAP_RETURN_INVERTED if inverted else aerospike.MAP_RETURN_NONE}

        if ctx is not None:
            self._fixed[_Keys.CTX_KEY] = ctx


class MapRemoveByKeyRelIndexRange(_BaseExpr):
    """Create an expression that removes map items nearest to key and greater by index with a count limit."""

    _op = aerospike.OP_MAP_REMOVE_BY_KEY_REL_INDEX_RANGE

    def __init__(
        self,
        ctx: "TypeCTX",
        key: "TypeKey",
        index: "TypeIndex",
        count: "TypeCount",
        bin: "TypeBinName",
        inverted: bool = False
    ):
        """Args:
            ctx (TypeCTX): An optional list of nested CDT :mod:`cdt_ctx <aerospike_helpers.cdt_ctx>` context operation
                objects.
            key (TypeKey): Key value or expression for key to start removing from.
            index (TypeIndex): Index integer or integer expression.
            count (TypeCount): Integer expression for how many elements to remove.
            bin (TypeBinName): bin expression, such as :class:`~aerospike_helpers.expressions.base.MapBin` or
                :class:`~aerospike_helpers.expressions.base.ListBin`.
            inverted (bool): Invert the expression's search criteria.

        :return: Map expression.

        Example::

            # Remove the next two items after key1
            # {"key1": 1, "key2": 2, "key3": 3, "key4": 10}
            expr = exp.MapRemoveByKeyRelIndexRange(None, "key1", 1, 2, exp.MapBin("b")).compile()
            # {"key1": 1, "key4": 10}
        """
        self._children = (
            key,
            index,
            count,
            bin if isinstance(bin, _BaseExpr) else MapBin(bin),
        )
        self._fixed = {_Keys.RETURN_TYPE_KEY: ReturnType.MAP_RETURN_INVERTED if inverted else aerospike.MAP_RETURN_NONE}

        if ctx is not None:
            self._fixed[_Keys.CTX_KEY] = ctx


class MapRemoveByValue(_BaseExpr):
    """Create an expression that removes map items identified by value."""

    _op = aerospike.OP_MAP_REMOVE_BY_VALUE

    def __init__(self, ctx: "TypeCTX", value: "TypeValue", bin: "TypeBinName", inverted: bool = False):
        """Args:
            ctx (TypeCTX): An optional list of nested CDT :mod:`cdt_ctx <aerospike_helpers.cdt_ctx>` context operation
                objects.
            value (TypeValue): Value or value expression to remove.
            bin (TypeBinName): bin expression, such as :class:`~aerospike_helpers.expressions.base.MapBin` or
                :class:`~aerospike_helpers.expressions.base.ListBin`.
            inverted (bool): Invert the expression's search criteria.

        :return: Map expression.

        Example::

            # Remove {"key1": 1} from map bin "b".
            expr = exp.MapRemoveByValue(None, 1, exp.MapBin("b")).compile()
        """
        self._children = (
            value,
            bin if isinstance(bin, _BaseExpr) else MapBin(bin),
        )
        self._fixed = {_Keys.RETURN_TYPE_KEY: ReturnType.MAP_RETURN_INVERTED if inverted else aerospike.MAP_RETURN_NONE}

        if ctx is not None:
            self._fixed[_Keys.CTX_KEY] = ctx


class MapRemoveByValueList(_BaseExpr):
    """Create an expression that removes map items identified by values."""

    _op = aerospike.OP_MAP_REMOVE_BY_VALUE_LIST

    def __init__(self, ctx: "TypeCTX", values: "TypeListValue", bin: "TypeBinName", inverted: bool = False):
        """Args:
            ctx (TypeCTX): An optional list of nested CDT :mod:`cdt_ctx <aerospike_helpers.cdt_ctx>` context operation
                objects.
            values (TypeListValue): List of values or list expression.
            bin (TypeBinName): bin expression, such as :class:`~aerospike_helpers.expressions.base.MapBin` or
                :class:`~aerospike_helpers.expressions.base.ListBin`.
            inverted (bool): Invert the expression's search criteria.

        :return: Map expression.

        Example::

            # Remove elements with values 1, 2, 3 from map bin "b".
            expr = exp.MapRemoveByValueList(None, [1, 2, 3], exp.MapBin("b")).compile()
        """
        self._children = (
            values,
            bin if isinstance(bin, _BaseExpr) else MapBin(bin),
        )
        self._fixed = {_Keys.RETURN_TYPE_KEY: ReturnType.MAP_RETURN_INVERTED if inverted else aerospike.MAP_RETURN_NONE}

        if ctx is not None:
            self._fixed[_Keys.CTX_KEY] = ctx


class MapRemoveByValueRange(_BaseExpr):
    """Create an expression that removes map items identified by value range
    (begin inclusive, end exclusive). If begin is nil, the range is less than end.
    If end is aerospike.CDTInfinite(), the range is greater than equal to begin.
    """

    _op = aerospike.OP_MAP_REMOVE_BY_VALUE_RANGE

    def __init__(
        self,
        ctx: "TypeCTX",
        begin: "TypeValue",
        end: "TypeValue",
        bin: "TypeBinName",
        inverted: bool = False
    ):
        """Args:
            ctx (TypeCTX): An optional list of nested CDT :mod:`cdt_ctx <aerospike_helpers.cdt_ctx>` context operation
                objects.
            begin (TypeValue): Begin value or value expression for range.
            end (TypeValue): End value or value expression for range.
            bin (TypeBinName): bin expression, such as :class:`~aerospike_helpers.expressions.base.MapBin` or
                :class:`~aerospike_helpers.expressions.base.ListBin`.
            inverted (bool): Invert the expression's search criteria.

        :return: Map expression.

        Example::

            # Remove list of items with values >= 3 and < 7 from map bin "b".
            expr = exp.MapRemoveByValueRange(None, 3, 7, exp.MapBin("b")).compile()
        """
        self._children = (
            begin,
            end,
            bin if isinstance(bin, _BaseExpr) else MapBin(bin),
        )
        self._fixed = {_Keys.RETURN_TYPE_KEY: ReturnType.MAP_RETURN_INVERTED if inverted else aerospike.MAP_RETURN_NONE}

        if ctx is not None:
            self._fixed[_Keys.CTX_KEY] = ctx


class MapRemoveByValueRelRankRangeToEnd(_BaseExpr):
    """Create an expression that removes map items nearest to value and greater by relative rank."""

    _op = aerospike.OP_MAP_REMOVE_BY_VALUE_REL_RANK_RANGE_TO_END

    def __init__(
        self,
        ctx: "TypeCTX",
        value: "TypeValue",
        rank: "TypeRank",
        bin: "TypeBinName",
        inverted: bool = False
    ):
        """Args:
            ctx (TypeCTX): An optional list of nested CDT :mod:`cdt_ctx <aerospike_helpers.cdt_ctx>` context operation
                objects.
            value (TypeValue): Value or value expression to start removing from.
            rank (TypeRank): Integer or integer expression of rank.
            bin (TypeBinName): bin expression, such as :class:`~aerospike_helpers.expressions.base.MapBin` or
                :class:`~aerospike_helpers.expressions.base.ListBin`.

        :return: Map expression.

        Example::

            # Remove all elements with values larger than 3 from map bin "b".
            expr = exp.MapRemoveByValueRelRankRangeToEnd(None, 3, 1, exp.MapBin("b")).compile()
        """
        self._children = (
            value,
            rank,
            bin if isinstance(bin, _BaseExpr) else MapBin(bin),
        )
        self._fixed = {_Keys.RETURN_TYPE_KEY: ReturnType.MAP_RETURN_INVERTED if inverted else aerospike.MAP_RETURN_NONE}

        if ctx is not None:
            self._fixed[_Keys.CTX_KEY] = ctx


class MapRemoveByValueRelRankRange(_BaseExpr):
    """Create an expression that removes map items nearest to value and greater by relative rank with a
    count limit.
    """

    _op = aerospike.OP_MAP_REMOVE_BY_VALUE_REL_RANK_RANGE

    def __init__(
        self,
        ctx: "TypeCTX",
        value: "TypeValue",
        rank: "TypeRank",
        count: "TypeCount",
        bin: "TypeBinName",
        inverted: bool = False
    ):
        """Args:
            ctx (TypeCTX): An optional list of nested CDT :mod:`cdt_ctx <aerospike_helpers.cdt_ctx>` context operation
                objects.
            value (TypeValue): Value or value expression to start removing from.
            rank (TypeRank): Integer or integer expression of rank.
            count (TypeCount): Integer count or integer expression for how many elements to remove.
            bin (TypeBinName): bin expression, such as :class:`~aerospike_helpers.expressions.base.MapBin` or
                :class:`~aerospike_helpers.expressions.base.ListBin`.
            inverted (bool): Invert the expression's search criteria.

        :return: Map expression.

        Example::

            # Remove the key with a value just lower than 17
            expr = exp.MapRemoveByValueRelRankRange(None, 17, -1, 1, exp.MapBin("b")).compile()
        """
        self._children = (
            value,
            rank,
            count,
            bin if isinstance(bin, _BaseExpr) else MapBin(bin),
        )
        self._fixed = {_Keys.RETURN_TYPE_KEY: ReturnType.MAP_RETURN_INVERTED if inverted else aerospike.MAP_RETURN_NONE}

        if ctx is not None:
            self._fixed[_Keys.CTX_KEY] = ctx


class MapRemoveByIndex(_BaseExpr):
    """Create an expression that removes map item identified by index."""

    _op = aerospike.OP_MAP_REMOVE_BY_INDEX

    def __init__(self, ctx: "TypeCTX", index: "TypeIndex", bin: "TypeBinName"):
        """Args:
            ctx (TypeCTX): An optional list of nested CDT :mod:`cdt_ctx <aerospike_helpers.cdt_ctx>` context operation
                objects.
            index (TypeIndex): Index integer or integer expression of element to remove.
            bin (TypeBinName): bin expression, such as :class:`~aerospike_helpers.expressions.base.MapBin` or
                :class:`~aerospike_helpers.expressions.base.ListBin`.

        :return: Map expression.

        Example::

            # Remove element with smallest key from map bin "b".
            expr = exp.MapRemoveByIndex(None, 0, exp.MapBin("b")).compile()
        """
        self._children = (
            index,
            bin if isinstance(bin, _BaseExpr) else MapBin(bin),
        )
        self._fixed = {}

        if ctx is not None:
            self._fixed[_Keys.CTX_KEY] = ctx


class MapRemoveByIndexRangeToEnd(_BaseExpr):
    """Create an expression that removes map items starting at specified index to the end of map."""

    _op = aerospike.OP_MAP_REMOVE_BY_INDEX_RANGE_TO_END

    def __init__(self, ctx: "TypeCTX", index: "TypeIndex", bin: "TypeBinName", inverted: bool = False):
        """Args:
            ctx (TypeCTX): An optional list of nested CDT :mod:`cdt_ctx <aerospike_helpers.cdt_ctx>` context operation
                objects.
            index (TypeIndex): Starting index integer or integer expression of elements to remove.
            bin (TypeBinName): bin expression, such as :class:`~aerospike_helpers.expressions.base.MapBin` or
                :class:`~aerospike_helpers.expressions.base.ListBin`.
            inverted (bool): Invert the expression's search criteria.

        :return: Map expression.

        Example::

            # Remove all elements starting from index 3 in map bin "b".
            expr = exp.MapRemoveByIndexRangeToEnd(None, 3, exp.MapBin("b")).compile()
        """
        self._children = (
            index,
            bin if isinstance(bin, _BaseExpr) else MapBin(bin),
        )
        self._fixed = {_Keys.RETURN_TYPE_KEY: ReturnType.MAP_RETURN_INVERTED if inverted else aerospike.MAP_RETURN_NONE}

        if ctx is not None:
            self._fixed[_Keys.CTX_KEY] = ctx


class MapRemoveByIndexRange(_BaseExpr):
    """Create an expression that removes count map items starting at specified index."""

    _op = aerospike.OP_MAP_REMOVE_BY_INDEX_RANGE

    def __init__(
            self,
            ctx: "TypeCTX",
            index: "TypeIndex",
            count: "TypeCount",
            bin: "TypeBinName",
            inverted: bool = False
    ):
        """Args:
            ctx (TypeCTX): An optional list of nested CDT :mod:`cdt_ctx <aerospike_helpers.cdt_ctx>` context operation
                objects.
            index (TypeIndex): Starting index integer or integer expression of elements to remove.
            count (TypeCount): Integer or integer expression, how many elements to remove.
            bin (TypeBinName): bin expression, such as :class:`~aerospike_helpers.expressions.base.MapBin` or
                :class:`~aerospike_helpers.expressions.base.ListBin`.
            inverted (bool): Invert the expression's search criteria.

        :return: Map expression.

        Example::

            # Get size of map bin "b" after index 3, 4, and 5 have been removed.
            expr = exp.MapSize(None, exp.MapRemoveByIndexRange(None, 3, 3, exp.MapBin("b"))).compile()
        """
        self._children = (
            index,
            count,
            bin if isinstance(bin, _BaseExpr) else MapBin(bin),
        )
        self._fixed = {_Keys.RETURN_TYPE_KEY: ReturnType.MAP_RETURN_INVERTED if inverted else aerospike.MAP_RETURN_NONE}

        if ctx is not None:
            self._fixed[_Keys.CTX_KEY] = ctx


class MapRemoveByRank(_BaseExpr):
    """Create an expression that removes map item identified by its value's rank."""

    _op = aerospike.OP_MAP_REMOVE_BY_RANK

    def __init__(self, ctx: "TypeCTX", rank: "TypeRank", bin: "TypeBinName"):
        """Args:
            ctx (TypeCTX): An optional list of nested CDT :mod:`cdt_ctx <aerospike_helpers.cdt_ctx>` context operation
                objects.
            rank (TypeRank): Rank integer or integer expression of element to remove.
            bin (TypeBinName): bin expression, such as :class:`~aerospike_helpers.expressions.base.MapBin` or
                :class:`~aerospike_helpers.expressions.base.ListBin`.

        :return: Map expression.

        Example::

            # Remove key with smallest value in map bin "b".
            expr = exp.MapRemoveByRank(None, 0, exp.MapBin("b")).compile()
        """
        self._children = (
            rank,
            bin if isinstance(bin, _BaseExpr) else MapBin(bin),
        )
        self._fixed = {}

        if ctx is not None:
            self._fixed[_Keys.CTX_KEY] = ctx


class MapRemoveByRankRangeToEnd(_BaseExpr):
    """Create an expression that removes map items starting at specified rank to the last ranked item."""

    _op = aerospike.OP_MAP_REMOVE_BY_RANK_RANGE_TO_END

    def __init__(self, ctx: "TypeCTX", rank: "TypeRank", bin: "TypeBinName", inverted: bool = False):
        """Args:
            ctx (TypeCTX): An optional list of nested CDT :mod:`cdt_ctx <aerospike_helpers.cdt_ctx>` context operation
                objects.
            rank (TypeRank): Rank integer or integer expression of element to start removing at.
            bin (TypeBinName): bin expression, such as :class:`~aerospike_helpers.expressions.base.MapBin` or
                :class:`~aerospike_helpers.expressions.base.ListBin`.
            inverted (bool): Invert the expression's search criteria.

        :return: Map expression.

        Example::

            # Remove keys with 2 largest values from map bin "b".
            expr = exp.MapRemoveByRankRangeToEnd(None, -2, exp.MapBin("b")).compile()
        """
        self._children = (
            rank,
            bin if isinstance(bin, _BaseExpr) else MapBin(bin),
        )
        self._fixed = {_Keys.RETURN_TYPE_KEY: ReturnType.MAP_RETURN_INVERTED if inverted else aerospike.MAP_RETURN_NONE}

        if ctx is not None:
            self._fixed[_Keys.CTX_KEY] = ctx


class MapRemoveByRankRange(_BaseExpr):
    """Create an expression that removes "count" map items starting at specified rank."""

    _op = aerospike.OP_MAP_REMOVE_BY_RANK_RANGE

    def __init__(
            self,
            ctx: "TypeCTX",
            rank: "TypeRank",
            count: "TypeCount",
            bin: "TypeBinName",
            inverted: bool = False
    ):
        """Args:
            ctx (TypeCTX): An optional list of nested CDT :mod:`cdt_ctx <aerospike_helpers.cdt_ctx>` context operation
                objects.
            rank (TypeRank): Rank integer or integer expression of element to start removing at.
            count (TypeCount): Count integer or integer expression of elements to remove.
            bin (TypeBinName): bin expression, such as :class:`~aerospike_helpers.expressions.base.MapBin` or
                :class:`~aerospike_helpers.expressions.base.ListBin`.
            inverted (bool): Invert the expression's search criteria.

        :return: Map expression.

        Example::

            # Remove 3 keys with the smallest values from map bin "b".
            expr = exp.MapRemoveByRankRange(None, 0, 3, exp.MapBin("b")).compile()
        """
        self._children = (
            rank,
            count,
            bin if isinstance(bin, _BaseExpr) else MapBin(bin),
        )
        self._fixed = {_Keys.RETURN_TYPE_KEY: ReturnType.MAP_RETURN_INVERTED if inverted else aerospike.MAP_RETURN_NONE}

        if ctx is not None:
            self._fixed[_Keys.CTX_KEY] = ctx


######################
# Map Read Expressions
######################


class MapSize(_BaseExpr):
    """Create an expression that returns map size."""

    _op = aerospike.OP_MAP_SIZE

    def __init__(self, ctx: "TypeCTX", bin: "TypeBinName"):
        """Args:
            ctx (TypeCTX): An optional list of nested CDT :mod:`cdt_ctx <aerospike_helpers.cdt_ctx>` context operation
                objects.
            bin (TypeBinName): bin expression, such as :class:`~aerospike_helpers.expressions.base.MapBin` or
                :class:`~aerospike_helpers.expressions.base.ListBin`.

        :return: Integer expression.

        Example::

            #Take the size of map bin "b".
            expr = exp.MapSize(None, exp.MapBin("b")).compile()
        """
        self._children = (bin if isinstance(bin, _BaseExpr) else MapBin(bin),)
        self._fixed = {}

        if ctx is not None:
            self._fixed[_Keys.CTX_KEY] = ctx


class MapGetByKey(_BaseExpr):
    """Create an expression that selects map item identified by key
    and returns selected data specified by return_type.
    """

    _op = aerospike.OP_MAP_GET_BY_KEY

    def __init__(self, ctx: "TypeCTX", return_type: int, value_type: int, key: "TypeKey", bin: "TypeBinName"):
        """Args:
            ctx (TypeCTX): An optional list of nested CDT :mod:`cdt_ctx <aerospike_helpers.cdt_ctx>` context operation
                objects.
            return_type (int): Value specifying what should be returned from the operation.
                This should be one of the :ref:`map_return_types` values.
            value_type (int): The value type that will be returned by this expression (ResultType).
            key (TypeKey): Key value or value expression of element to get.
            bin (TypeBinName): bin expression, such as :class:`~aerospike_helpers.expressions.base.MapBin` or
                :class:`~aerospike_helpers.expressions.base.ListBin`.

        :return: Expression.

        Example::

            # Get the value at key "key0" in map bin "b". (assume the value at key0 is an integer)
            expr = exp.MapGetByKey(None, aerospike.MAP_RETURN_VALUE, exp.ResultType.INTEGER, "key0",
                exp.MapBin("b")).compile()
        """
        self._children = (
            key,
            bin if isinstance(bin, _BaseExpr) else MapBin(bin),
        )
        self._fixed = {_Keys.VALUE_TYPE_KEY: value_type, _Keys.RETURN_TYPE_KEY: return_type}

        if ctx is not None:
            self._fixed[_Keys.CTX_KEY] = ctx


class MapGetByKeyRange(_BaseExpr):
    """Create an expression that selects map items identified by key range.
    (begin inclusive, end exclusive). If begin is nil, the range is less than end.
    If end is aerospike.CDTInfinite(), the range is greater than equal to begin.
    Expression returns selected data specified by return_type.
    """

    _op = aerospike.OP_MAP_GET_BY_KEY_RANGE

    def __init__(
        self,
        ctx: "TypeCTX",
        return_type: int,
        begin: "TypeKey",
        end: "TypeKey",
        bin: "TypeBinName",
        inverted: bool = False
    ):
        """Args:
            ctx (TypeCTX): An optional list of nested CDT :mod:`cdt_ctx <aerospike_helpers.cdt_ctx>` context operation
                objects.
            return_type (int): Value specifying what should be returned from the operation.
                This should be one of the :ref:`map_return_types` values.
            begin (TypeKey): Key value or expression.
            end (TypeKey): Key value or expression.
            bin (TypeBinName): bin expression, such as :class:`~aerospike_helpers.expressions.base.MapBin` or
                :class:`~aerospike_helpers.expressions.base.ListBin`.
            inverted (bool): Invert the expression's search criteria.

        :return: Expression.

        Example::

            # Get elements at keys "key3", "key4", "key5", "key6" in map bin "b".
            expr = exp.MapGetByKeyRange(None, aerospike.MAP_RETURN_VALUE, "key3", "key7", exp.MapBin("b")).compile()
        """
        self._children = (
            begin,
            end,
            bin if isinstance(bin, _BaseExpr) else MapBin(bin),
        )
        self._fixed = {_Keys.RETURN_TYPE_KEY: return_type}
        if inverted:
            self._fixed[_Keys.RETURN_TYPE_KEY] |= ReturnType.MAP_RETURN_INVERTED

        if ctx is not None:
            self._fixed[_Keys.CTX_KEY] = ctx


class MapGetByKeyList(_BaseExpr):
    """Create an expression that selects map items identified by keys
    and returns selected data specified by return_type.
    """

    _op = aerospike.OP_MAP_GET_BY_KEY_LIST

    def __init__(
        self,
        ctx: "TypeCTX",
        return_type: int,
        keys: "TypeKeyList",
        bin: "TypeBinName",
        inverted: bool = False
    ):
        """Args:
            ctx (TypeCTX): An optional list of nested CDT :mod:`cdt_ctx <aerospike_helpers.cdt_ctx>` context operation
                objects.
            return_type (int): Value specifying what should be returned from the operation.
                This should be one of the :ref:`map_return_types` values.
            keys (TypeKeyList): List of key values or list expression.
            bin (TypeBinName): bin expression, such as :class:`~aerospike_helpers.expressions.base.MapBin` or
                :class:`~aerospike_helpers.expressions.base.ListBin`.
            inverted (bool): Invert the expression's search criteria.

        :return: Expression.

        Example::

            # Get elements at keys "key3", "key4", "key5" in map bin "b".
            expr = exp.MapGetByKeyList(None, aerospike.MAP_RETURN_VALUE, ["key3", "key4", "key5"],
                exp.MapBin("b")).compile()
        """
        self._children = (
            keys,
            bin if isinstance(bin, _BaseExpr) else MapBin(bin),
        )
        self._fixed = {_Keys.RETURN_TYPE_KEY: return_type}
        if inverted:
            self._fixed[_Keys.RETURN_TYPE_KEY] |= ReturnType.MAP_RETURN_INVERTED

        if ctx is not None:
            self._fixed[_Keys.CTX_KEY] = ctx


class MapGetByKeyRelIndexRangeToEnd(_BaseExpr):
    """Create an expression that selects map items nearest to key and greater by index with a count limit.
    Expression returns selected data specified by return_type.
    """

    _op = aerospike.OP_MAP_GET_BY_KEY_REL_INDEX_RANGE_TO_END

    def __init__(
        self,
        ctx: "TypeCTX",
        return_type: int,
        key: "TypeKey",
        index: "TypeIndex",
        bin: "TypeBinName",
        inverted: bool = False
    ):
        """Args:
            ctx (TypeCTX): An optional list of nested CDT :mod:`cdt_ctx <aerospike_helpers.cdt_ctx>` context operation
                objects.
            return_type (int): Value specifying what should be returned from the operation.
                This should be one of the :ref:`map_return_types` values.
            key (TypeKey): Key value or value expression.
            index (TypeIndex): Index integer or integer value expression.
            bin (TypeBinName): bin expression, such as :class:`~aerospike_helpers.expressions.base.MapBin` or
                :class:`~aerospike_helpers.expressions.base.ListBin`.
            inverted (bool): Invert the expression's search criteria.

        :return: Expression.

        Example::

            # Get elements with keys larger than "key2" from map bin "b".
            expr = exp.MapGetByKeyRelIndexRangeToEnd(None, aerospike.MAP_RETURN_VALUE, "key2", 1,
                exp.MapBin("b")).compile()
        """
        self._children = (
            key,
            index,
            bin if isinstance(bin, _BaseExpr) else MapBin(bin),
        )
        self._fixed = {_Keys.RETURN_TYPE_KEY: return_type}
        if inverted:
            self._fixed[_Keys.RETURN_TYPE_KEY] |= ReturnType.MAP_RETURN_INVERTED

        if ctx is not None:
            self._fixed[_Keys.CTX_KEY] = ctx


class MapGetByKeyRelIndexRange(_BaseExpr):
    """Create an expression that selects map items nearest to key and greater by index with a count limit.
    Expression returns selected data specified by return_type.
    """

    _op = aerospike.OP_MAP_GET_BY_KEY_REL_INDEX_RANGE

    def __init__(
        self,
        ctx: "TypeCTX",
        return_type: int,
        key: "TypeKey",
        index: "TypeIndex",
        count: "TypeCount",
        bin: "TypeBinName",
        inverted: bool = False
    ):
        """Args:
            ctx (TypeCTX): An optional list of nested CDT :mod:`cdt_ctx <aerospike_helpers.cdt_ctx>` context operation
                objects.
            return_type (int): Value specifying what should be returned from the operation.
                This should be one of the :ref:`map_return_types` values.
            key (TypeKey): Key value or value expression.
            index (TypeIndex): Index integer or integer value expression.
            count (TypeCount): Integer count or integer value expression.
            bin (TypeBinName): bin expression, such as :class:`~aerospike_helpers.expressions.base.MapBin` or
                :class:`~aerospike_helpers.expressions.base.ListBin`.
            inverted (bool): Invert the expression's search criteria.

        :return: Expression.

        Example::

            expr = exp.MapGetByKeyRelIndexRange(None, aerospike.MAP_RETURN_VALUE, "key2", 0, 2,
                exp.MapBin("b")).compile()
            # [2, 3]
            expr = exp.MapGetByKeyRelIndexRange(None, aerospike.MAP_RETURN_VALUE, "key2", 1, 2,
                exp.MapBin("b")).compile()
            # [3, 10]

        """
        self._children = (
            key,
            index,
            count,
            bin if isinstance(bin, _BaseExpr) else MapBin(bin),
        )
        self._fixed = {_Keys.RETURN_TYPE_KEY: return_type}
        if inverted:
            self._fixed[_Keys.RETURN_TYPE_KEY] |= ReturnType.MAP_RETURN_INVERTED

        if ctx is not None:
            self._fixed[_Keys.CTX_KEY] = ctx


class MapGetByValue(_BaseExpr):
    """Create an expression that selects map items identified by value
    and returns selected data specified by return_type.
    """

    _op = aerospike.OP_MAP_GET_BY_VALUE

    def __init__(
        self,
        ctx: "TypeCTX",
        return_type: int,
        value: "TypeValue",
        bin: "TypeBinName",
        inverted: bool = False
    ):
        """Args:
            ctx (TypeCTX): An optional list of nested CDT :mod:`cdt_ctx <aerospike_helpers.cdt_ctx>` context operation
                objects.
            return_type (int): Value specifying what should be returned from the operation.
                This should be one of the :ref:`map_return_types` values.
            value (TypeValue): Value or value expression of element to get.
            bin (TypeBinName): bin expression, such as :class:`~aerospike_helpers.expressions.base.MapBin` or
                :class:`~aerospike_helpers.expressions.base.ListBin`.
            inverted (bool): Invert the expression's search criteria.

        :return: Expression.

        Example::

            # Get the rank of the element with value, 3, in map bin "b".
            expr = exp.MapGetByValue(None, aerospike.MAP_RETURN_RANK, 3, exp.MapBin("b")).compile()
        """
        self._children = (value, bin if isinstance(bin, _BaseExpr) else MapBin(bin))
        self._fixed = {_Keys.RETURN_TYPE_KEY: return_type}
        if inverted:
            self._fixed[_Keys.RETURN_TYPE_KEY] |= ReturnType.MAP_RETURN_INVERTED

        if ctx is not None:
            self._fixed[_Keys.CTX_KEY] = ctx


class MapGetByValueRange(_BaseExpr):
    """Create an expression that selects map items identified by value range.
    (begin inclusive, end exclusive). If begin is None, the range is less than end.
    If end is None, the range is greater than equal to begin.
    Expression returns selected data specified by return_type.
    """

    _op = aerospike.OP_MAP_GET_BY_VALUE_RANGE

    def __init__(
        self,
        ctx: "TypeCTX",
        return_type: int,
        value_begin: "TypeValue",
        value_end: "TypeValue",
        bin: "TypeBinName",
        inverted: bool = False
    ):
        """Args:
            ctx (TypeCTX): An optional list of nested CDT :mod:`cdt_ctx <aerospike_helpers.cdt_ctx>` context operation
                objects.
            return_type (int): Value specifying what should be returned from the operation.
                This should be one of the :ref:`map_return_types` values.
            value_begin (TypeValue): Value or value expression of first element to get.
            value_end (TypeValue): Value or value expression of ending element.
            bin (TypeBinName): bin expression, such as :class:`~aerospike_helpers.expressions.base.MapBin` or
                :class:`~aerospike_helpers.expressions.base.ListBin`.
            inverted (bool): Invert the expression's search criteria.

        :return: Expression.

        Example::

            # Get elements with values between 3 and 7 from map bin "b".
            expr = exp.MapGetByValueRange(None, aerospike.MAP_RETURN_VALUE, 3, 7, exp.MapBin("b")).compile()
        """
        self._children = (value_begin, value_end, bin if isinstance(bin, _BaseExpr) else MapBin(bin))
        self._fixed = {_Keys.RETURN_TYPE_KEY: return_type}
        if inverted:
            self._fixed[_Keys.RETURN_TYPE_KEY] |= ReturnType.MAP_RETURN_INVERTED

        if ctx is not None:
            self._fixed[_Keys.CTX_KEY] = ctx


class MapGetByValueList(_BaseExpr):
    """Create an expression that selects map items identified by values
    and returns selected data specified by return_type.
    """

    _op = aerospike.OP_MAP_GET_BY_VALUE_LIST

    def __init__(
        self,
        ctx: "TypeCTX",
        return_type: int,
        value: "TypeListValue",
        bin: "TypeBinName",
        inverted: bool = False
    ):
        """Args:
            ctx (TypeCTX): An optional list of nested CDT :mod:`cdt_ctx <aerospike_helpers.cdt_ctx>` context operation
                objects.
            return_type (int): Value specifying what should be returned from the operation.
                This should be one of the :ref:`map_return_types` values.
            value (TypeListValue): List or list expression of values of elements to get.
            bin (TypeBinName): bin expression, such as :class:`~aerospike_helpers.expressions.base.MapBin` or
                :class:`~aerospike_helpers.expressions.base.ListBin`.
            inverted (bool): Invert the expression's search criteria.

        :return: Expression.

        Example::

            # Get the indexes of the the elements in map bin "b" with values [3, 6, 12].
            expr = exp.MapGetByValueList(None, aerospike.MAP_RETURN_INDEX, [3, 6, 12], exp.MapBin("b")).compile()
        """
        self._children = (value, bin if isinstance(bin, _BaseExpr) else MapBin(bin))
        self._fixed = {_Keys.RETURN_TYPE_KEY: return_type}
        if inverted:
            self._fixed[_Keys.RETURN_TYPE_KEY] |= ReturnType.MAP_RETURN_INVERTED

        if ctx is not None:
            self._fixed[_Keys.CTX_KEY] = ctx


class MapGetByValueRelRankRangeToEnd(_BaseExpr):
    """Create an expression that selects map items nearest to value and greater by relative rank,
    Expression returns selected data specified by return_type.
    """

    _op = aerospike.OP_MAP_GET_BY_VALUE_RANK_RANGE_REL_TO_END

    def __init__(
        self,
        ctx: "TypeCTX",
        return_type: int,
        value: "TypeValue",
        rank: "TypeRank",
        bin: "TypeBinName",
        inverted: bool = False
    ):
        """Args:
            ctx (TypeCTX): An optional list of nested CDT :mod:`cdt_ctx <aerospike_helpers.cdt_ctx>` context operation
                objects.
            return_type (int): Value specifying what should be returned from the operation.
                This should be one of the :ref:`map_return_types` values.
            value (TypeValue): Value or value expression to get items relative to.
            rank (TypeRank): Rank integer expression. rank relative to "value" to start getting elements.
            bin (TypeBinName): bin expression, such as :class:`~aerospike_helpers.expressions.base.MapBin` or
                :class:`~aerospike_helpers.expressions.base.ListBin`.
            inverted (bool): Invert the expression's search criteria.

        :return: Expression.

        Example::

            # Get the values of all elements in map bin "b" larger than 3.
            expr = exp.MapGetByValueRelRankRangeToEnd(None, aerospike.MAP_RETURN_VALUE, 3, 1, exp.MapBin("b")).compile()
        """
        self._children = (value, rank, bin if isinstance(bin, _BaseExpr) else MapBin(bin))
        self._fixed = {_Keys.RETURN_TYPE_KEY: return_type}
        if inverted:
            self._fixed[_Keys.RETURN_TYPE_KEY] |= ReturnType.MAP_RETURN_INVERTED

        if ctx is not None:
            self._fixed[_Keys.CTX_KEY] = ctx


class MapGetByValueRelRankRange(_BaseExpr):
    """Create an expression that selects map items nearest to value and greater by relative rank with a
    count limit. Expression returns selected data specified by return_type.
    """

    _op = aerospike.OP_MAP_GET_BY_VALUE_RANK_RANGE_REL

    def __init__(
        self,
        ctx: "TypeCTX",
        return_type: int,
        value: "TypeValue",
        rank: "TypeRank",
        count: "TypeCount",
        bin: "TypeBinName",
        inverted: bool = False
    ):
        """Args:
            ctx (TypeCTX): An optional list of nested CDT :mod:`cdt_ctx <aerospike_helpers.cdt_ctx>` context operation
                objects.
            return_type (int): Value specifying what should be returned from the operation.
                This should be one of the :ref:`map_return_types` values.
            value (TypeValue): Value or value expression to get items relative to.
            rank (TypeRank): Rank integer expression. rank relative to "value" to start getting elements.
            count (TypeCount): Integer value or integer value expression, how many elements to get.
            bin (TypeBinName): bin expression, such as :class:`~aerospike_helpers.expressions.base.MapBin` or
                :class:`~aerospike_helpers.expressions.base.ListBin`.
            inverted (bool): Invert the expression's search criteria.

        :return: Expression.

        Example::

            # {"key1": 1, "key2": 2, "key3": 3, "key4": 10}
            # Get next two largest values greater than a value of 1
            expr = exp.MapGetByValueRelRankRange(None, aerospike.MAP_RETURN_VALUE, 1, 1, 2, exp.MapBin("b")).compile()
            # [2, 3]
        """
        self._children = (value, rank, count, bin if isinstance(bin, _BaseExpr) else MapBin(bin))
        self._fixed = {_Keys.RETURN_TYPE_KEY: return_type}
        if inverted:
            self._fixed[_Keys.RETURN_TYPE_KEY] |= ReturnType.MAP_RETURN_INVERTED

        if ctx is not None:
            self._fixed[_Keys.CTX_KEY] = ctx


class MapGetByIndex(_BaseExpr):
    """Create an expression that selects map item identified by index
    and returns selected data specified by return_type.
    """

    _op = aerospike.OP_MAP_GET_BY_INDEX

    def __init__(
        self,
        ctx: "TypeCTX",
        return_type: int,
        value_type: int,
        index: "TypeIndex",
        bin: "TypeBinName",
    ):
        """Args:
            ctx (TypeCTX): An optional list of nested CDT :mod:`cdt_ctx <aerospike_helpers.cdt_ctx>` context operation
                objects.
            return_type (int): Value specifying what should be returned from the operation.
                This should be one of the :ref:`map_return_types` values.
            value_type (int): The value type that will be returned by this expression (ResultType).
            index (TypeIndex): Integer or integer expression of index to get element at.
            bin (TypeBinName): bin expression, such as :class:`~aerospike_helpers.expressions.base.MapBin` or
                :class:`~aerospike_helpers.expressions.base.ListBin`.

        :return: Expression.

        Example::

            # Get the value at index 0 in map bin "b". (assume this value is an integer)
            expr = exp.MapGetByIndex(None, aerospike.MAP_RETURN_VALUE,
                                     exp.ResultType.INTEGER, 0, exp.MapBin("b")).compile()
        """
        self._children = (index, bin if isinstance(bin, _BaseExpr) else MapBin(bin))
        self._fixed = {_Keys.VALUE_TYPE_KEY: value_type, _Keys.RETURN_TYPE_KEY: return_type}

        if ctx is not None:
            self._fixed[_Keys.CTX_KEY] = ctx


class MapGetByIndexRangeToEnd(_BaseExpr):
    """Create an expression that selects map items starting at specified index to the end of map
    and returns selected data specified by return_type.
    """

    _op = aerospike.OP_MAP_GET_BY_INDEX_RANGE_TO_END

    def __init__(
        self,
        ctx: "TypeCTX",
        return_type: int,
        index: "TypeIndex",
        bin: "TypeBinName",
        inverted: bool = False
    ):
        """Args:
            ctx (TypeCTX): An optional list of nested CDT :mod:`cdt_ctx <aerospike_helpers.cdt_ctx>` context operation
                objects.
            return_type (int): Value specifying what should be returned from the operation.
                This should be one of the :ref:`map_return_types` values.
            index (TypeIndex): Integer or integer expression of index to start getting elements at.
            bin (TypeBinName): bin expression, such as :class:`~aerospike_helpers.expressions.base.MapBin` or
                :class:`~aerospike_helpers.expressions.base.ListBin`.
            inverted (bool): Invert the expression's search criteria.

        :return: Expression.

        Example::

            # Get element at index 5 to end from map bin "b".
            expr = exp.MapGetByIndexRangeToEnd(None, aerospike.MAP_RETURN_VALUE, 5, exp.MapBin("b")).compile()
        """
        self._children = (index, bin if isinstance(bin, _BaseExpr) else MapBin(bin))
        self._fixed = {_Keys.RETURN_TYPE_KEY: return_type}
        if inverted:
            self._fixed[_Keys.RETURN_TYPE_KEY] |= ReturnType.MAP_RETURN_INVERTED

        if ctx is not None:
            self._fixed[_Keys.CTX_KEY] = ctx


class MapGetByIndexRange(_BaseExpr):
    """Create an expression that selects "count" map items starting at specified index
    and returns selected data specified by return_type.
    """

    _op = aerospike.OP_MAP_GET_BY_INDEX_RANGE

    def __init__(
        self,
        ctx: "TypeCTX",
        return_type: int,
        index: "TypeIndex",
        count: "TypeCount",
        bin: "TypeBinName",
        inverted: bool = False
    ):
        """Args:
            ctx (TypeCTX): An optional list of nested CDT :mod:`cdt_ctx <aerospike_helpers.cdt_ctx>` context operation
                objects.
            return_type (int): Value specifying what should be returned from the operation.
                This should be one of the :ref:`map_return_types` values.
            index (TypeIndex): Integer or integer expression of index to start getting elements at.
            count (TypeCount): Integer or integer expression for count of elements to get.
            bin (TypeBinName): bin expression, such as :class:`~aerospike_helpers.expressions.base.MapBin` or
                :class:`~aerospike_helpers.expressions.base.ListBin`.
            inverted (bool): Invert the expression's search criteria.

        :return: Expression.

        Example::

            # Get elements at indexes 3, 4, 5, 6 in map bin "b".
            expr = exp.MapGetByIndexRange(None, aerospike.MAP_RETURN_VALUE, 3, 4, exp.MapBin("b")).compile()
        """
        self._children = (index, count, bin if isinstance(bin, _BaseExpr) else MapBin(bin))
        self._fixed = {_Keys.RETURN_TYPE_KEY: return_type}
        if inverted:
            self._fixed[_Keys.RETURN_TYPE_KEY] |= ReturnType.MAP_RETURN_INVERTED

        if ctx is not None:
            self._fixed[_Keys.CTX_KEY] = ctx


class MapGetByRank(_BaseExpr):
    """Create an expression that selects map items identified by rank
    and returns selected data specified by return_type.
    """

    _op = aerospike.OP_MAP_GET_BY_RANK

    def __init__(
        self,
        ctx: "TypeCTX",
        return_type: int,
        value_type: int,
        rank: "TypeRank",
        bin: "TypeBinName",
    ):
        """Args:
            ctx (TypeCTX): An optional list of nested CDT :mod:`cdt_ctx <aerospike_helpers.cdt_ctx>` context operation
                objects.
            return_type (int): Value specifying what should be returned from the operation.
                This should be one of the :ref:`map_return_types` values.
            value_type (int): The value type that will be returned by this expression (ResultType).
            rank (TypeRank): Rank integer or integer expression of element to get.
            bin (TypeBinName): bin expression, such as :class:`~aerospike_helpers.expressions.base.MapBin` or
                :class:`~aerospike_helpers.expressions.base.ListBin`.

        :return: Expression.

        Example::

            # Get the smallest element in map bin "b".
            expr = exp.MapGetByRank(None, aerospike.MAP_RETURN_VALUE, exp.ResultType.INTEGER, 0,
                exp.MapBin("b")).compile()
        """
        self._children = (rank, bin if isinstance(bin, _BaseExpr) else MapBin(bin))
        self._fixed = {_Keys.VALUE_TYPE_KEY: value_type, _Keys.RETURN_TYPE_KEY: return_type}

        if ctx is not None:
            self._fixed[_Keys.CTX_KEY] = ctx


class MapGetByRankRangeToEnd(_BaseExpr):
    """Create an expression that selects map items starting at specified rank to the last ranked item
    and returns selected data specified by return_type.
    """

    _op = aerospike.OP_MAP_GET_BY_RANK_RANGE_TO_END

    def __init__(self, ctx: "TypeCTX", return_type: int, rank: "TypeRank", bin: "TypeBinName", inverted: bool = False):
        """Args:
            ctx (TypeCTX): An optional list of nested CDT :mod:`cdt_ctx <aerospike_helpers.cdt_ctx>` context operation
                objects.
            return_type (int): Value specifying what should be returned from the operation.
                This should be one of the :ref:`map_return_types` values.
            rank (TypeRank): Rank integer or integer expression of first element to get.
            bin (TypeBinName): bin expression, such as :class:`~aerospike_helpers.expressions.base.MapBin` or
                :class:`~aerospike_helpers.expressions.base.ListBin`.
            inverted (bool): Invert the expression's search criteria.

        :return: Expression.

        Example::

            # Get the three largest elements in map bin "b".
            expr = exp.MapGetByRankRangeToEnd(None, aerospike.MAP_RETURN_VALUE, -3, MapBin("b")).compile()
        """
        self._children = (rank, bin if isinstance(bin, _BaseExpr) else MapBin(bin))
        self._fixed = {_Keys.RETURN_TYPE_KEY: return_type}
        if inverted:
            self._fixed[_Keys.RETURN_TYPE_KEY] |= ReturnType.MAP_RETURN_INVERTED

        if ctx is not None:
            self._fixed[_Keys.CTX_KEY] = ctx


class MapGetByRankRange(_BaseExpr):
    """Create an expression that selects "count" map items starting at specified rank
    and returns selected data specified by return_type.
    """

    _op = aerospike.OP_MAP_GET_BY_RANK_RANGE

    def __init__(
        self,
        ctx: "TypeCTX",
        return_type: int,
        rank: "TypeRank",
        count: "TypeCount",
        bin: "TypeBinName",
        inverted: bool = False
    ):
        """Args:
            ctx (TypeCTX): An optional list of nested CDT :mod:`cdt_ctx <aerospike_helpers.cdt_ctx>` context operation
                objects.
            return_type (int): Value specifying what should be returned from the operation.
                This should be one of the :ref:`map_return_types` values.
            rank (TypeRank): Rank integer or integer expression of first element to get.
            count (TypeCount): Count integer or integer expression for how many elements to get.
            bin (TypeBinName): bin expression, such as :class:`~aerospike_helpers.expressions.base.MapBin` or
                :class:`~aerospike_helpers.expressions.base.ListBin`.
            inverted (bool): Invert the expression's search criteria.

        :return: Expression.

        Example::

            # Get the 3 smallest elements in map bin "b".
            expr = exp.MapGetByRankRange(None, aerospike.MAP_RETURN_VALUE, 0, 3, exp.MapBin("b")).compile()
        """
        self._children = (rank, count, bin if isinstance(bin, _BaseExpr) else MapBin(bin))
        self._fixed = {_Keys.RETURN_TYPE_KEY: return_type}
        if inverted:
            self._fixed[_Keys.RETURN_TYPE_KEY] |= ReturnType.MAP_RETURN_INVERTED

        if ctx is not None:
            self._fixed[_Keys.CTX_KEY] = ctx



from typing import Union

def between(bin: str, min: int, max: int) -> tuple: ...
def equals(bin: str, val: Union[str, int]) -> tuple: ...

def geo_within_geojson_region(bin: str, shape: str, index_type = ...) -> tuple: ...
def geo_within_radius(bin: str, long: float, lat: float, radius_meters: float, index_type = ...) -> tuple: ...
def geo_contains_geojson_point(bin: str, point: str, index_type = ...) -> tuple: ...
def geo_contains_point(bin: str, long: float, lat: float, index_type = ...) -> tuple: ...

def contains(bin: str, index_type, val = Union[str, int]) -> tuple: ...
def range(bin: str, index_type, min: int, max: int) -> tuple: ...



##########################################################################
# Copyright 2013-2021 Aerospike, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##########################################################################

class HyperLogLog(bytes):
    """
    Represents a HyperLogLog value. This can be returned from the server or created in order to be sent to the server.

    The constructor takes in any argument that the :class:`bytes` constructor takes in.

    >>> h = HyperLogLog([1, 2, 3])
    >>> client.put(key, {"hyperloglog": h})
    """
    def __new__(cls, o) -> "HyperLogLog":
        return super().__new__(cls, o)

    # We need to implement repr() and str() ourselves
    # Otherwise, this class will inherit these methods from bytes
    # making it indistinguishable from bytes objects when printed
    def __repr__(self) -> str:
        bytes_str = super().__repr__()
        return f"{self.__class__.__name__}({bytes_str})"

    def __str__(self) -> str:
        return self.__repr__()



##########################################################################
# Copyright 2013-2021 Aerospike, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##########################################################################
"""
Helper functions to generate complex data type context (cdt_ctx) objects for use with operations on nested CDTs (list,
map, etc).

Example::

    import aerospike
    from aerospike import exception as ex
    from aerospike_helpers import cdt_ctx
    from aerospike_helpers.operations import map_operations
    from aerospike_helpers.operations import list_operations
    import sys

    # Configure the client.
    config = {"hosts": [("127.0.0.1", 3000)]}
    client = aerospike.client(config)

    key = ("test", "demo", "foo")
    listWithMaps = [
        {"name": "John", "id": 100},
        {"name": "Bill", "id": 200}
    ]
    binName = "users"

    # Write the record
    client.put(key, {binName: listWithMaps})

    # Example 1: read the id of the second person on the list
    # Get context of the second person
    ctx = [cdt_ctx.cdt_ctx_list_index(1)]
    ops = [
        map_operations.map_get_by_key(
            binName, "id", aerospike.MAP_RETURN_VALUE, ctx
        )
    ]

    _, _, result = client.operate(key, ops)
    print(result)
    # {'users': 200}

    # Example 2: add a new person and get their rating of Facebook
    cindy = {
        "name": "Cindy",
        "id": 300,
        "ratings": {
            "Facebook": 4,
            "Snapchat": 5
        }
    }

    # Context list used for read operation after adding Cindy
    # Cindy will be the third person (index 2)
    # Then go to their ratings
    ctx = [cdt_ctx.cdt_ctx_list_index(2), cdt_ctx.cdt_ctx_map_key("ratings")]
    ops = [
        list_operations.list_append(binName, cindy),
        map_operations.map_get_by_key(
            binName, "Facebook", aerospike.MAP_RETURN_VALUE, ctx
        )
    ]

    _, _, result = client.operate(key, ops)
    print(result)
    # {'users': 4}

    # Example 3: create a CDT secondary index from a base64 encoded _cdt_ctx with info command
    policy = {}

    bs_b4_cdt = client.get_cdtctx_base64(ctx_list_index)

    r = []
    r.append("sindex-create:ns=test;set=demo;indexname=test_string_list_cdt_index")
    # use index_type_string to convert enum value to string
    r.append(";indextype=%s" % (cdt_ctx.index_type_string(aerospike.INDEX_TYPE_LIST)))
    # use index_datatype_string to convert enum value to string
    r.append(";indexdata=string_list,%s" % (cdt_ctx.index_datatype_string(aerospike.INDEX_STRING)))
    r.append(";context=%s" % (bs_b4_cdt))
    req = ''.join(r)

    # print("req is ==========={}", req)
    retobj = client.info_all(req, policy=None)
    # print("res is ==========={}", res)
    client.index_remove('test', 'test_string_list_cdt_index', policy)

    # Cleanup
    client.remove(key)
    client.close()
"""
import aerospike


def index_type_string(index_type):
    """
    Converts index_type enum value to string.

    Args:
        index_type (int): The index_type to convert into equivalent string value.

    Returns:
        (string) - must be one of 'default', 'list', 'mapkeys', 'mapvalues'

    """
    if index_type == aerospike.INDEX_TYPE_DEFAULT:
        return "default"
    if index_type == aerospike.INDEX_TYPE_LIST:
        return "list"
    if index_type == aerospike.INDEX_TYPE_MAPKEYS:
        return "mapkeys"
    if index_type == aerospike.INDEX_TYPE_MAPVALUES:
        return "mapvalues"
    return "invalid"


def index_datatype_string(index_datatype):
    """
    Converts index_datatype enum value to string.

    Args:
        index_datatype (int): The index_datatype to convert into equivalent string value.

    Returns:
        (string) - must be one of must be one of 'numeric', 'string', 'geo2dsphere'
    """
    if index_datatype == aerospike.INDEX_NUMERIC:
        return "numeric"
    if index_datatype == aerospike.INDEX_STRING:
        return "string"
    if index_datatype == aerospike.INDEX_GEO2DSPHERE:
        return "geo2dsphere"
    return "invalid"


CDT_CTX_ORDER_KEY = "order_key"
CDT_CTX_PAD_KEY = "pad_key"


class _cdt_ctx:
    """
    Class used to represent a single ctx_operation.
    """

    def __init__(self, *, id=None, value=None, extra_args=None):
        self.id = id
        self.value = value
        self.extra_args = extra_args


def cdt_ctx_list_index(index):
    """
    Creates a nested cdt_ctx object to lookup an object in a list by index.

    If the index is negative, the lookup starts backwards from the end of the list.
    If it is out of bounds, a parameter error will be returned.

    Args:
        index (int): The index to look for in the list.

    Returns:
        :class:`~aerospike_helpers.cdt_ctx._cdt_ctx`
    """
    return _cdt_ctx(id=aerospike.CDT_CTX_LIST_INDEX, value=index)


def cdt_ctx_list_rank(rank):
    """
    Creates a nested cdt_ctx object to lookup an object in a list by rank.

    If the rank is negative, the lookup starts backwards from the largest rank value.

    Args:
        rank (int): The rank to look for in the list.

    Returns:
        :class:`~aerospike_helpers.cdt_ctx._cdt_ctx`
    """
    return _cdt_ctx(id=aerospike.CDT_CTX_LIST_RANK, value=rank)


def cdt_ctx_list_value(value):
    """
    Creates a nested cdt_ctx object to lookup an object in a list by value.

    Args:
        value (object): The value to look for in the list.

    Returns:
        :class:`~aerospike_helpers.cdt_ctx._cdt_ctx`
    """
    return _cdt_ctx(id=aerospike.CDT_CTX_LIST_VALUE, value=value)


def cdt_ctx_list_index_create(index: int, order: int = 0, pad: bool = False) -> _cdt_ctx:
    """
    Creates a nested cdt_ctx object to create an list and insert at a given index.

    If a list already exists at the index, a new list will not be created.
    Any operations using this cdt_ctx object will be applied to the existing list.

    If a non-list element exists at the index, an :py:exc:`~aerospike.exception.InvalidRequest` will be thrown.

    Args:
        index (int): The index to create the list at.
        order (int): The :ref:`sort order <aerospike_list_order>` to create the List with.
            (default: ``aerospike.LIST_UNORDERED``)
        pad (bool): If index is out of bounds and ``pad`` is :py:obj:`True`,
            then the list will be created at the index with :py:obj:`None` elements inserted behind it.
            ``pad`` is only compatible with unordered lists.

    Returns:
        :class:`~aerospike_helpers.cdt_ctx._cdt_ctx`
    """
    return _cdt_ctx(
        id=aerospike.CDT_CTX_LIST_INDEX_CREATE, value=index, extra_args={CDT_CTX_ORDER_KEY: order, CDT_CTX_PAD_KEY: pad}
    )


def cdt_ctx_map_index(index):
    """
    The cdt_ctx object is initialized to lookup an object in a map by index.

    If the index is negative, the lookup starts backwards from the end of the map.

    If it is out of bounds, a parameter error will be returned.

    Args:
        index (int): The index to look for in the map.

    Returns:
        :class:`~aerospike_helpers.cdt_ctx._cdt_ctx`
    """
    return _cdt_ctx(id=aerospike.CDT_CTX_MAP_INDEX, value=index)


def cdt_ctx_map_rank(rank):
    """
    The cdt_ctx object is initialized to lookup an object in a map by index.

    If the rank is negative, the lookup starts backwards from the largest rank value.

    Args:
        rank (int): The rank to look for in the map.

    Returns:
        :class:`~aerospike_helpers.cdt_ctx._cdt_ctx`
    """
    return _cdt_ctx(id=aerospike.CDT_CTX_MAP_RANK, value=rank)


def cdt_ctx_map_key(key):
    """
    The cdt_ctx object is initialized to lookup an object in a map by key.

    Args:
        key (object): The key to look for in the map.

    Returns:
        :class:`~aerospike_helpers.cdt_ctx._cdt_ctx`
    """
    return _cdt_ctx(id=aerospike.CDT_CTX_MAP_KEY, value=key)


def cdt_ctx_map_value(value):
    """
    The cdt_ctx object is initialized to lookup an object in a map by value.

    Args:
        value (object): The value to look for in the map.

    Returns:
        :class:`~aerospike_helpers.cdt_ctx._cdt_ctx`
    """
    return _cdt_ctx(id=aerospike.CDT_CTX_MAP_VALUE, value=value)


def cdt_ctx_map_key_create(key: any, order: int = 0) -> _cdt_ctx:
    """
    Create a map with the given sort order at the given key.

    Args:
        key (object): The key to create the map at.
        order (int): The :ref:`sort order <aerospike_map_order>` to create the List with.
            (default: ``aerospike.MAP_UNORDERED``)

    Returns:
        :class:`~aerospike_helpers.cdt_ctx._cdt_ctx`
    """
    return _cdt_ctx(id=aerospike.CDT_CTX_MAP_KEY_CREATE, value=key, extra_args={CDT_CTX_ORDER_KEY: order})



##########################################################################
# Copyright 2013-2021 Aerospike, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##########################################################################
"""
Arithmetic expressions provide arithmetic operator support for Aerospike expressions.

"""

# from __future__ import annotations
from typing import Union

from aerospike_helpers.expressions.resources import _GenericExpr
from aerospike_helpers.expressions.resources import _BaseExpr
from aerospike_helpers.expressions.resources import _ExprOp

TypeNumber = Union[_BaseExpr, int, float]
TypeFloat = Union[_BaseExpr, float]
TypeInteger = Union[_BaseExpr, int]


########################
# Arithmetic Expressions
########################


class Add(_BaseExpr):
    """Create an add, (+) expression.
    All arguments must be the same type (integer or float).

    Add is also available via operator overloading using `+`
    and any subclass of _BaseExpr. See the second example.

    Requires server version 5.6.0+.
    """

    _op = _ExprOp.ADD

    def __init__(self, *args: "TypeNumber"):
        """Args:
            `*args` (TypeNumber): Variable amount of float or integer expressions or values to be added together.

        :return: (integer or float value).

        Example::

            # Integer bin "a" + "b" == 11
            expr = exp.Eq(exp.Add(exp.IntBin("a"), exp.IntBin("b")), 11).compile()

            # Using operator overloading.
            expr = exp.Eq(exp.IntBin("a") + exp.IntBin("b"), 11).compile()

        """
        self._children = args + (_GenericExpr(_ExprOp._AS_EXP_CODE_END_OF_VA_ARGS, 0, {}),)


class Sub(_BaseExpr):
    """
    Create "subtract" (-) operator that applies to a variable number of expressions.
    If only one argument is provided, return the negation of that argument.
    Otherwise, return the sum of the 2nd to Nth argument subtracted from the 1st
    argument. All arguments must resolve to the same type (integer or float).

    Sub is also available via operator overloading using `-`
    and any subclass of _BaseExpr. See the second example.

    Requires server version 5.6.0+.
    """

    _op = _ExprOp.SUB

    def __init__(self, *args: "TypeNumber"):
        """Args:
            `*args` (TypeNumber): Variable amount of float or integer expressions or values to be subtracted.

        :return: (integer or float value)

        Example::

            # Integer bin "a" - "b" == 11
            expr = exp.Eq(exp.Sub(exp.IntBin("a"), exp.IntBin("b")), 11).compile()

            # Using operator overloading.
            expr = exp.Eq(exp.IntBin("a") - exp.IntBin("b"), 11).compile()
        """
        self._children = args + (_GenericExpr(_ExprOp._AS_EXP_CODE_END_OF_VA_ARGS, 0, {}),)


class Mul(_BaseExpr):
    """
    Create "multiply" (*) operator that applies to a variable number of expressions.
    Return the product of all arguments. If only one argument is supplied, return
    that argument. All arguments must resolve to the same type (integer or float).

    Mul is also available via operator overloading using `*`
    and any subclass of _BaseExpr. See the second example.

    Requires server version 5.6.0+.
    """

    _op = _ExprOp.MUL

    def __init__(self, *args: "TypeNumber"):
        """Args:
            `*args` (TypeNumber): Variable amount of float or integer expressions or values to be multiplied.

        :return: (integer or float value)

        Example::

            # Integer bin "a" * "b" >= 11
            expr = exp.GE(exp.Mul(exp.IntBin("a"), exp.IntBin("b")), 11).compile()

            # Using operator overloading.
            expr = exp.GE(exp.IntBin("a") * exp.IntBin("b"), 11).compile()
        """
        self._children = args + (_GenericExpr(_ExprOp._AS_EXP_CODE_END_OF_VA_ARGS, 0, {}),)


class Div(_BaseExpr):
    """Create "divide" (/) operator that applies to a variable number of expressions.
    If there is only one argument, returns the reciprocal for that argument.
    Otherwise, return the first argument divided by the product of the rest.
    All arguments must resolve to the same type (integer or float).

    Div is also available via operator overloading using `/`
    and any subclass of _BaseExpr. See the second example.

    Floor div is also available via `//` but must be used with floats.

    Requires server version 5.6.0+.
    """

    _op = _ExprOp.DIV

    def __init__(self, *args: "TypeNumber"):
        """Args:
            `*args` (TypeNumber): Variable amount of float or integer expressions or values to be divided.

        :return: (integer or float value)

        Example::

            # Integer bin "a" / "b" / "c" >= 11
            expr = exp.GE(exp.Div(exp.IntBin("a"), exp.IntBin("b"), exp.IntBin("c")), 11).compile()

            # Using operator overloading.
            expr = exp.GE(exp.IntBin("a") / exp.IntBin("b") / exp.IntBin("c"), 11).compile()

            # Float bin "a" // "b" // "c" >= 11.0
            expr = exp.GE(exp.FloatBin("a") // exp.FloatBin("b") // exp.FloatBin("c"), 11.0).compile()
        """
        self._children = args + (_GenericExpr(_ExprOp._AS_EXP_CODE_END_OF_VA_ARGS, 0, {}),)


class Pow(_BaseExpr):
    """
    Create "pow" operator that raises a "base" to the "exponent" power.
    All arguments must resolve to floats.

    Pow is also available via operator overloading using `**`
    and any subclass of _BaseExpr. See the second example.

    Requires server version 5.6.0+.
    """

    _op = _ExprOp.POW

    def __init__(self, base: "TypeFloat", exponent: "TypeFloat"):
        """Args:
            base (TypeFloat): Float expression or value base.
            exponent (TypeFloat): Float expression or value exponent.

        :return: (float value)

        Example::

            # Float bin "a" ** 2.0 == 16.0
            expr = exp.Eq(exp.Pow(exp.FloatBin("a"), 2.0), 16.0).compile()

            # Using operator overloading.
            expr = exp.Eq(exp.FloatBin("a") ** 2.0, 16.0).compile()
        """
        self._children = (base, exponent)


class Log(_BaseExpr):
    """
    Create "log" operator for logarithm of "num" with base "base".
    All arguments must resolve to floats.

    Requires server version 5.6.0+.
    """

    _op = _ExprOp.LOG

    def __init__(self, num: "TypeFloat", base: "TypeFloat"):
        """Args:
            num (TypeFloat): Float expression or value number.
            base (TypeFloat): Float expression or value base.

        :return: (float value)

        Example::

            # For float bin "a", log("a", 2.0) == 16.0
            expr = exp.Eq(exp.Log(exp.FloatBin("a"), 2.0), 16.0).compile()
        """
        self._children = (num, base)


class Mod(_BaseExpr):
    """
    Create "modulo" (%) operator that determines the remainder of "numerator"
    divided by "denominator". All arguments must resolve to integers.

    Mod is also available via operator overloading using `%`
    and any subclass of _BaseExpr. See the second example.

    Requires server version 5.6.0+.
    """

    _op = _ExprOp.MOD

    def __init__(self, numerator: "TypeInteger", denominator: "TypeInteger"):
        """Args:
            numerator (TypeInteger): Integer expression or value numerator.
            denominator (TypeInteger): Integer expression or value denominator.

        :return: (integer value)

        Example::

            # For int bin "a" % 10 == 0
            expr = exp.Eq(exp.Mod(exp.IntBin("a"), 10), 0).compile()

            # Using operator overloading.
            expr = exp.Eq(exp.IntBin("a") % 10, 0).compile()
        """
        self._children = (numerator, denominator)


class Abs(_BaseExpr):
    """
    Create operator that returns absolute value of a number.
    All arguments must resolve to integer or float.

    Abs is also available via operator overloading using the built-in
    abs() function and any subclass of _BaseExpr. See the second example.

    Requires server version 5.6.0+.
    """

    _op = _ExprOp.ABS

    def __init__(self, value: "TypeNumber"):
        """Args:
            value (TypeNumber): Float or integer expression or value to take absolute value of.

        :return: (number value)

        Example::

            # For int bin "a", abs("a") == 1
            expr = exp.Eq(exp.Abs(exp.IntBin("a")), 1).compile()

            # Using operator overloading
            expr = exp.Eq(abs(exp.IntBin("a")), 1).compile()
        """
        self._children = (value,)


class Floor(_BaseExpr):
    """
    Create floor expression that rounds a floating point number down
    to the closest integer value.

    Floor is also available via operator overloading using the math.floor()
    function and any subclass of _BaseExpr. See the second example.

    Requires server version 5.6.0+.
    """

    _op = _ExprOp.FLOOR

    def __init__(self, value: "TypeFloat"):
        """Args:
            value (TypeFloat): Float expression or value to take floor of.

        :return: (float value)

        Example::

            # Floor(2.25) == 2.0
            expr = exp.Eq(exp.Floor(2.25), 2.0).compile()

            # Using operator overloading
            import math
            expr = exp.Eq(math.floor(2.25), 2.0).compile()
        """
        self._children = (value,)


class Ceil(_BaseExpr):
    """
    Create ceil expression that rounds a floating point number up
    to the closest integer value.

    Ceil is also available via operator overloading using the math.ceil()
    function and any subclass of _BaseExpr. See the second example.

    Requires server version 5.6.0+.
    """

    _op = _ExprOp.CEIL

    def __init__(self, value: "TypeFloat"):
        """Args:
            value (TypeFloat): Float expression or value to take ceiling of.

        :return: (float value)

        Example::

            # Ceil(2.25) == 3.0
            expr = exp.Eq(exp.Ceil(2.25), 3.0).compile()

            # Using operator overloading
            import math
            expr = exp.Eq(math.ceil(2.25), 3.0).compile()
        """
        self._children = (value,)


class ToInt(_BaseExpr):
    """Create expression that converts a float to an integer.

    Requires server version 5.6.0+.
    """

    _op = _ExprOp.TO_INT

    def __init__(self, value: "TypeFloat"):
        """Args:
            value (TypeFloat): Float expression or value to convert to int.

        :return: (integer value)

        Example::

            #For float bin "a", int(exp.FloatBin("a")) == 2
            expr = exp.Eq(exp.ToInt(exp.FloatBin("a")), 2).compile()
        """
        self._children = (value,)


class ToFloat(_BaseExpr):
    """Create expression that converts an integer to a float.

    Requires server version 5.6.0+.
    """

    _op = _ExprOp.TO_FLOAT

    def __init__(self, value: "TypeInteger"):
        """Args:
            value (TypeInteger): Integer expression or value to convert to float.

        :return: (float value)

        Example::

            #For int bin "a", float(exp.IntBin("a")) == 2
            expr = exp.Eq(exp.ToFloat(exp.IntBin("a")), 2).compile()
        """
        self._children = (value,)


class Min(_BaseExpr):
    """
    Create expression that returns the minimum value in a variable number of expressions.
    All arguments must be the same type (integer or float).

    Requires server version 5.6.0+.
    """

    _op = _ExprOp.MIN

    def __init__(self, *args: "TypeNumber"):
        """Args:
            `*args` (TypeNumber): Variable amount of float or integer expressions or values from which to find the
            minimum value.

        :return: (integer or float value).

        Example::

            # for integer bins a, b, c, min(a, b, c) > 0
            expr = exp.GT(exp.Min(exp.IntBin("a"), exp.IntBin("b"), exp.IntBin("c")), 0).compile()
        """
        self._children = args + (_GenericExpr(_ExprOp._AS_EXP_CODE_END_OF_VA_ARGS, 0, {}),)


class Max(_BaseExpr):
    """Create expression that returns the maximum value in a variable number of expressions.
    All arguments must be the same type (integer or float).

    Requires server version 5.6.0+.
    """

    _op = _ExprOp.MAX

    def __init__(self, *args: "TypeNumber"):
        """Args:
            `*args` (TypeNumber): Variable amount of float or integer expressions or values from which to find the
            maximum value.

        :return: (integer or float value).

        Example::

            # for integer bins a, b, c, max(a, b, c) > 100
            expr = exp.GT(exp.Max(exp.IntBin("a"), exp.IntBin("b"), exp.IntBin("c")), 100).compile()
        """
        self._children = args + (_GenericExpr(_ExprOp._AS_EXP_CODE_END_OF_VA_ARGS, 0, {}),)



##########################################################################
# Copyright 2013-2024 Aerospike, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##########################################################################

"""Classes used for metrics.

:class:`ConnectionStats`, :class:`NodeMetrics`, :class:`Node`, and :class:`Cluster` do not have a constructor
because they are not meant to be created by the user. They are only meant to be returned from :class:`MetricsListeners`
callbacks for reading data about the server and client.
"""

from typing import Optional, Callable


class ConnectionStats:
    """Connection statistics.

    Attributes:
        in_use (int): Connections actively being used in database transactions on this node.
            There can be multiple pools per node. This value is a summary of those pools on this node.
        in_pool (int): Connections residing in pool(s) on this node.
            There can be multiple pools per node. This value is a summary of those pools on this node.
        opened (int): Total number of node connections opened since node creation.
        closed (int): Total number of node connections closed since node creation.
    """
    pass


class NodeMetrics:
    """
    Each type of latency has a list of latency buckets.

    Latency bucket counts are cumulative and not reset on each metrics snapshot interval.

    Attributes:
        conn_latency (list[int])
        write_latency (list[int])
        read_latency (list[int])
        batch_latency (list[int])
        query_latency (list[int])
    """
    pass


class Node:
    """Server node representation.

    Attributes:
        name (str): The name of the node.
        address (str): The IP address / host name of the node (not including the port number).
        port (int): Port number of the node's address.
        conns (:py:class:`ConnectionStats`): Synchronous connection stats on this node.
        error_count (int): Transaction error count since node was initialized. If the error is retryable,
            multiple errors per transaction may occur.
        timeout_count (int): Transaction timeout count since node was initialized.
            If the timeout is retryable (ie socketTimeout), multiple timeouts per transaction may occur.
        metrics (:py:class:`NodeMetrics`): Node metrics
    """
    pass


class Cluster:
    """Cluster of server nodes.

    Attributes:
        cluster_name (Optional[str]): Expected cluster name for all nodes. May be :py:obj:`None`.
        invalid_node_count (int): Count of add node failures in the most recent cluster tend iteration.
        tran_count (int): Transaction count. The value is cumulative and not reset per metrics interval.
        retry_count (int): Transaction retry count. There can be multiple retries for a single transaction.
            The value is cumulative and not reset per metrics interval.
        nodes (list[:py:class:`Node`]): Active nodes in cluster.
    """
    pass


class MetricsListeners:
    """Metrics listener callbacks.

    All callbacks must be set.

    Attributes:
        enable_listener (Callable[[], None]): Periodic extended metrics has been enabled for the given cluster.
        snapshot_listener (Callable[[Cluster], None]): A metrics snapshot has been requested for the given cluster.
        node_close_listener (Callable[[Node], None]): A node is being dropped from the cluster.
        disable_listener (Callable[[Cluster], None]): Periodic extended metrics has been disabled for the given cluster.
    """
    def __init__(
            self,
            enable_listener: Callable[[], None],
            snapshot_listener: Callable[[Cluster], None],
            node_close_listener: Callable[[Node], None],
            disable_listener: Callable[[Cluster], None]
    ):
        self.enable_listener = enable_listener
        self.snapshot_listener = snapshot_listener
        self.node_close_listener = node_close_listener
        self.disable_listener = disable_listener


class MetricsPolicy:
    """Client periodic metrics configuration.

    Attributes:
        metrics_listeners (Optional[:py:class:`MetricsListeners`]): Listeners that handles metrics notification events.
            If set to :py:obj:`None`, the default listener implementation will be used, which writes the metrics
            snapshot to a file which can later be read and forwarded to OpenTelemetry by a separate offline
            application. Otherwise, use all listeners set in the class instance.

            The listener could be overridden to send the metrics snapshot directly to OpenTelemetry.
        report_dir (str): Directory path to write metrics log files for listeners that write logs.
        report_size_limit (int): Metrics file size soft limit in bytes for listeners that write logs.
            When report_size_limit is reached or exceeded, the current metrics file is closed and a new
            metrics file is created with a new timestamp. If report_size_limit is zero, the metrics file
            size is unbounded and the file will only be closed when :py:meth:`~aerospike.Client.disable_metrics` or
            :py:meth:`~aerospike.Client.close()` is called.
        interval (int): Number of cluster tend iterations between metrics notification events. One tend iteration
            is defined as ``"tend_interval"`` in the client config plus the time to tend all nodes.
        latency_columns (int): Number of elapsed time range buckets in latency histograms.
        latency_shift (int): Power of 2 multiple between each range bucket in latency histograms starting at column 3.
            The bucket units are in milliseconds. The first 2 buckets are "<=1ms" and ">1ms".

            Example::

                # latencyColumns=7 latencyShift=1
                # <=1ms >1ms >2ms >4ms >8ms >16ms >32ms

                # latencyColumns=5 latencyShift=3
                # <=1ms >1ms >8ms >64ms >512ms
    """
    def __init__(
            self,
            metrics_listeners: Optional[MetricsListeners] = None,
            report_dir: str = ".",
            report_size_limit: int = 0,
            interval: int = 30,
            latency_columns: int = 7,
            latency_shift: int = 1
    ):
        self.metrics_listeners = metrics_listeners
        self.report_dir = report_dir
        self.report_size_limit = report_size_limit
        self.interval = interval
        self.latency_columns = latency_columns
        self.latency_shift = latency_shift



##########################################################################
# Copyright 2013-2021 Aerospike, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##########################################################################
"""
This module provides helper functions to produce dictionaries to be used with:

* :mod:`aerospike.Client.operate` and :mod:`aerospike.Client.operate_ordered`
* Certain batch operations listed in :mod:`aerospike_helpers.batch.records`

List operations support nested CDTs through an optional ctx context argument.
The ctx argument is a list of cdt_ctx context operation objects. See :class:`aerospike_helpers.cdt_ctx`.

For list operations that take in an index parameter, you can use negative values for indexes.
See this `page <https://aerospike.com/docs/server/guide/data-types/cdt-list#list-terminology>`_.

.. note:: Nested CDT (ctx) requires server version >= 4.6.0

"""
import aerospike
from typing import Optional


OP_KEY = "op"
BIN_KEY = "bin"
VALUE_KEY = "val"
INDEX_KEY = "index"
LIST_POLICY_KEY = "list_policy"
INVERTED_KEY = "inverted"
RETURN_TYPE_KEY = "return_type"
COUNT_KEY = "count"
RANK_KEY = "rank"
VALUE_BEGIN_KEY = "value_begin"
VALUE_END_KEY = "value_end"
VALUE_LIST_KEY = "value_list"
LIST_ORDER_KEY = "list_order"
SORT_FLAGS_KEY = "sort_flags"
CTX_KEY = "ctx"
PAD_KEY = "pad"
PERSIST_INDEX_KEY = "persist_index"


def list_create(bin_name: str, list_order: int, pad: bool, persist_index: bool, ctx: Optional[list] = None):
    """
    Create list create operation.

    Server creates list at given context level.

    Args:
        bin_name (str):	Bin name.
        list_order (int): See :ref:`aerospike_list_order` for possible values.
        persist_index (bool): If :py:obj:`True`, persist list index. A list index improves lookup performance,
            but requires more storage. A list index can be created for a top-level
            ordered list only. Nested and unordered list indexes are not supported.
        ctx (Optional[dict]): An optional list of nested CDT :class:`cdt_ctx <aerospike_helpers.cdt_ctx>`
            specifying the path to nested list. If not defined, the top-level list is used.
    """
    op_dict = {
        OP_KEY: aerospike.OP_LIST_CREATE,
        BIN_KEY: bin_name,
        LIST_ORDER_KEY: list_order,
        PAD_KEY: pad,
        PERSIST_INDEX_KEY: persist_index
    }

    if ctx is not None:
        op_dict[CTX_KEY] = ctx

    return op_dict


def list_append(bin_name: str, value, policy: Optional[dict] = None, ctx: Optional[list] = None):
    """
    Creates a list append operation.

    The list append operation instructs the aerospike server to append an item to the
    end of a list bin.

    Returns a dictionary that maps the bin name to the list size after appending the item.

    Args:
        bin_name (str): The name of the bin to be operated on.
        value: The value to be appended to the end of the list.
        policy (dict): An optional dictionary of :ref:`list write options <aerospike_list_policies>`.
        ctx (list): An optional list of nested CDT :class:`cdt_ctx <aerospike_helpers.cdt_ctx>` context operation
            objects.
    Returns:
        A dictionary usable in :meth:`~aerospike.Client.operate` and :meth:`~aerospike.Client.operate_ordered`.
        The format of the dictionary should be considered an internal detail, and subject to change.
    """
    op_dict = {OP_KEY: aerospike.OP_LIST_APPEND, BIN_KEY: bin_name, VALUE_KEY: value}

    if policy:
        op_dict[LIST_POLICY_KEY] = policy

    if ctx:
        op_dict[CTX_KEY] = ctx

    return op_dict


def list_append_items(bin_name: str, values, policy: Optional[dict] = None, ctx: Optional[list] = None):
    """
    Creates a list append items operation.

    The list append items operation instructs the aerospike server to append multiple items to the
    end of a list bin.

    Args:
        bin_name (str): The name of the bin to be operated on.
        values (list): A sequence of items to be appended to the end of the list.
        policy (dict): An optional dictionary of :ref:`list write options <aerospike_list_policies>`.
        ctx (list): An optional list of nested CDT :class:`cdt_ctx <aerospike_helpers.cdt_ctx>` context operation
            objects.
    Returns:
        A dictionary usable in :meth:`~aerospike.Client.operate` and :meth:`~aerospike.Client.operate_ordered`.
        The format of the dictionary should be considered an internal detail, and subject to change.
    """
    op_dict = {OP_KEY: aerospike.OP_LIST_APPEND_ITEMS, BIN_KEY: bin_name, VALUE_KEY: values}

    if policy:
        op_dict[LIST_POLICY_KEY] = policy

    if ctx:
        op_dict[CTX_KEY] = ctx

    return op_dict


def list_insert(bin_name: str, index, value, policy: Optional[dict] = None, ctx: Optional[list] = None):
    """
    Creates a list insert operation.

    The list insert operation inserts an item at index: `index` into the list contained
    in the specified bin.

    Args:
        bin_name (str): The name of the bin to be operated on.
        index (int): The index at which to insert an item. The value may be positive to use
            zero based indexing or negative to index from the end of the list.
        value: The value to be inserted into the list.
        policy (dict): An optional dictionary of :ref:`list write options <aerospike_list_policies>`.
        ctx (list): An optional list of nested CDT :class:`cdt_ctx <aerospike_helpers.cdt_ctx>` context operation
            objects.
    Returns:
        A dictionary usable in :meth:`~aerospike.Client.operate` and :meth:`~aerospike.Client.operate_ordered`.
        The format of the dictionary should be considered an internal detail, and subject to change.
    """
    op_dict = {OP_KEY: aerospike.OP_LIST_INSERT, BIN_KEY: bin_name, INDEX_KEY: index, VALUE_KEY: value}

    if policy:
        op_dict[LIST_POLICY_KEY] = policy

    if ctx:
        op_dict[CTX_KEY] = ctx

    return op_dict


def list_insert_items(bin_name: str, index, values, policy: Optional[dict] = None, ctx: Optional[list] = None):
    """
    Creates a list insert items operation.

    The list insert items operation inserts items at index: `index` into the list contained
    in the specified bin. Server returns list size.

    Args:
        bin_name (str): The name of the bin to be operated on.
        index (int): The index at which to insert the items. The value may be positive to use
            zero based indexing or negative to index from the end of the list.
        values (list): The values to be inserted into the list.
        policy (dict): An optional dictionary of :ref:`list write options <aerospike_list_policies>`.
        ctx (list): An optional list of nested CDT :class:`cdt_ctx <aerospike_helpers.cdt_ctx>` context operation
            objects.
    Returns:
        A dictionary usable in :meth:`~aerospike.Client.operate` and :meth:`~aerospike.Client.operate_ordered`.
        The format of the dictionary should be considered an internal detail, and subject to change.
    """
    op_dict = {OP_KEY: aerospike.OP_LIST_INSERT_ITEMS, BIN_KEY: bin_name, INDEX_KEY: index, VALUE_KEY: values}

    if policy:
        op_dict[LIST_POLICY_KEY] = policy

    if ctx:
        op_dict[CTX_KEY] = ctx

    return op_dict


def list_increment(bin_name: str, index, value, policy: Optional[dict] = None, ctx: Optional[list] = None):
    """
    Creates a list increment operation.

    The list increment operation increments an item at index: `index` in the list contained
    in the specified bin.

    Args:
        bin_name (str): The name of the bin to be operated on.
        index (int): The index of the list item to increment.
        value (int, float) : The value to be added to the list item.
        policy (dict): An optional dictionary of :ref:`list write options <aerospike_list_policies>`.
        ctx (list): An optional list of nested CDT :class:`cdt_ctx <aerospike_helpers.cdt_ctx>` context operation
            objects.
    Returns:
        A dictionary usable in :meth:`~aerospike.Client.operate` and :meth:`~aerospike.Client.operate_ordered`.
        The format of the dictionary should be considered an internal detail, and subject to change.
    """
    op_dict = {OP_KEY: aerospike.OP_LIST_INCREMENT, BIN_KEY: bin_name, INDEX_KEY: index, VALUE_KEY: value}

    if policy:
        op_dict[LIST_POLICY_KEY] = policy

    if ctx:
        op_dict[CTX_KEY] = ctx

    return op_dict


def list_pop(bin_name: str, index, ctx: Optional[list] = None):
    """
    Creates a list pop operation.

    The list pop operation removes and returns an item index: `index` from list contained
    in the specified bin.

    Args:
        bin_name (str): The name of the bin to be operated on.
        index (int): The index of the item to be removed.
        ctx (list): An optional list of nested CDT :class:`cdt_ctx <aerospike_helpers.cdt_ctx>` context operation
            objects.
    Returns:
        A dictionary usable in :meth:`~aerospike.Client.operate` and :meth:`~aerospike.Client.operate_ordered`. The
        format of the dictionary should be considered an internal detail, and subject to change.
    """
    op_dict = {OP_KEY: aerospike.OP_LIST_POP, BIN_KEY: bin_name, INDEX_KEY: index}

    if ctx:
        op_dict[CTX_KEY] = ctx

    return op_dict


def list_pop_range(bin_name: str, index, count, ctx: Optional[list] = None):
    """
    Creates a list pop range operation.

    The list pop range operation removes and returns `count` items
    starting from index: `index` from the list contained in the specified bin.

    Args:
        bin_name (str): The name of the bin to be operated on.
        index (int): The index of the first item to be removed.
        count (int): A positive number indicating how many items, including the first, to be removed and returned
        ctx (list): An optional list of nested CDT :class:`cdt_ctx <aerospike_helpers.cdt_ctx>` context operation
            objects.
    Returns:
        A dictionary usable in :meth:`~aerospike.Client.operate` and :meth:`~aerospike.Client.operate_ordered`. The
        format of the dictionary should be considered an internal detail, and subject to change.
    """
    op_dict = {OP_KEY: aerospike.OP_LIST_POP_RANGE, BIN_KEY: bin_name, INDEX_KEY: index, VALUE_KEY: count}

    if ctx:
        op_dict[CTX_KEY] = ctx

    return op_dict


def list_remove(bin_name: str, index, ctx: Optional[list] = None):
    """
    Creates a list remove operation.

    The list remove operation removes an item located at `index` in the list specified by `bin_name`.
    Server returns number of items removed.

    Args:
        bin_name (str): The name of the bin containing the item to be removed.
        index (int): The index at which to remove the item.
        ctx (list): An optional list of nested CDT :class:`cdt_ctx <aerospike_helpers.cdt_ctx>` context operation
            objects.
    Returns:
        A dictionary usable in :meth:`~aerospike.Client.operate` and :meth:`~aerospike.Client.operate_ordered`. The
        format of the dictionary should be considered an internal detail, and subject to change.
    """
    op_dict = {OP_KEY: aerospike.OP_LIST_REMOVE, BIN_KEY: bin_name, INDEX_KEY: index}

    if ctx:
        op_dict[CTX_KEY] = ctx

    return op_dict


def list_remove_range(bin_name: str, index, count, ctx: Optional[list] = None):
    """
    Creates a list remove range operation.

    The list remove range operation removes `count` items starting at `index`
    in the list specified by `bin_name`. Server returns number of items removed.

    Args:
        bin_name (str): The name of the bin containing the items to be removed.
        index (int): The index of the first item to remove.
        count (int): A positive number representing the number of items to be removed.
        ctx (list): An optional list of nested CDT :class:`cdt_ctx <aerospike_helpers.cdt_ctx>` context operation
            objects.
    Returns:
        A dictionary usable in :meth:`~aerospike.Client.operate` and :meth:`~aerospike.Client.operate_ordered`. The
        format of the dictionary should be considered an internal detail, and subject to change.
    """
    op_dict = {OP_KEY: aerospike.OP_LIST_REMOVE_RANGE, BIN_KEY: bin_name, INDEX_KEY: index, VALUE_KEY: count}

    if ctx:
        op_dict[CTX_KEY] = ctx

    return op_dict


def list_clear(bin_name: str, ctx: Optional[list] = None):
    """Create list clear operation.

    The list clear operation removes all items from the list specified by `bin_name`

    Args:
        bin_name (str): The name of the bin containing the list to be cleared
        ctx (list): An optional list of nested CDT :class:`cdt_ctx <aerospike_helpers.cdt_ctx>` context operation
            objects.

    Returns:
        A dictionary usable in :meth:`~aerospike.Client.operate` and :meth:`~aerospike.Client.operate_ordered`. The
        format of the dictionary should be considered an internal detail, and subject to change.
    """
    op_dict = {OP_KEY: aerospike.OP_LIST_CLEAR, BIN_KEY: bin_name}

    if ctx:
        op_dict[CTX_KEY] = ctx

    return op_dict


def list_set(bin_name: str, index, value, policy: Optional[dict] = None, ctx: Optional[list] = None):
    """Create a list set operation.

    The list set operations sets the value of the item at `index` to `value`.
    Server does not return a result by default.

    Args:
        bin_name (str): The name of the bin containing the list to be operated on.
        index (int): The index of the item to be set.
        value: The value to be assigned to the list item.
        policy (dict): An optional dictionary of :ref:`list write options <aerospike_list_policies>`.
        ctx (list): An optional list of nested CDT :class:`cdt_ctx <aerospike_helpers.cdt_ctx>` context operation
            objects.

    Returns:
        A dictionary usable in :meth:`~aerospike.Client.operate` and :meth:`~aerospike.Client.operate_ordered`. The
        format of the dictionary should be considered an internal detail, and subject to change.
    """
    op_dict = {OP_KEY: aerospike.OP_LIST_SET, BIN_KEY: bin_name, INDEX_KEY: index, VALUE_KEY: value}
    if policy:
        op_dict[LIST_POLICY_KEY] = policy

    if ctx:
        op_dict[CTX_KEY] = ctx

    return op_dict


def list_get(bin_name: str, index, ctx: Optional[list] = None):
    """Create a list get operation.

    The list get operation gets the value of the item at `index` and returns the value

    Args:
        bin_name (str): The name of the bin containing the list to fetch items from.
        index (int): The index of the item to be returned.
        ctx (list): An optional list of nested CDT :class:`cdt_ctx <aerospike_helpers.cdt_ctx>` context operation
            objects.

    Returns:
        A dictionary usable in :meth:`~aerospike.Client.operate` and :meth:`~aerospike.Client.operate_ordered`. The
        format of the dictionary should be considered an internal detail, and subject to change.
    """
    op_dict = {OP_KEY: aerospike.OP_LIST_GET, BIN_KEY: bin_name, INDEX_KEY: index}

    if ctx:
        op_dict[CTX_KEY] = ctx

    return op_dict


def list_get_range(bin_name: str, index, count, ctx: Optional[list] = None):
    """Create a list get range operation.

    The list get range operation gets `count` items starting `index` and returns the values.

    Args:
        bin_name (str): The name of the bin containing the list to fetch items from.
        index (int): The index of the item to be returned.
        count (int): A positive number of items to be returned.
        ctx (list): An optional list of nested CDT :class:`cdt_ctx <aerospike_helpers.cdt_ctx>` context operation
            objects.

    Returns:
        A dictionary usable in :meth:`~aerospike.Client.operate` and :meth:`~aerospike.Client.operate_ordered`. The
        format of the dictionary should be considered an internal detail, and subject to change.
    """
    op_dict = {OP_KEY: aerospike.OP_LIST_GET_RANGE, BIN_KEY: bin_name, INDEX_KEY: index, VALUE_KEY: count}

    if ctx:
        op_dict[CTX_KEY] = ctx

    return op_dict


def list_trim(bin_name: str, index, count, ctx: Optional[list] = None):
    """Create a list trim operation.

    Server removes items in list bin that do not fall into range specified by index and count range.

    Args:
        bin_name (str): The name of the bin containing the list to be trimmed.
        index (int): The index of the items to be kept.
        count (int): A positive number of items to be kept.
        ctx (list): An optional list of nested CDT :class:`cdt_ctx <aerospike_helpers.cdt_ctx>` context operation
            objects.

    Returns:
        A dictionary usable in :meth:`~aerospike.Client.operate` and :meth:`~aerospike.Client.operate_ordered`. The
        format of the dictionary should be considered an internal detail, and subject to change.
    """
    op_dict = {OP_KEY: aerospike.OP_LIST_TRIM, BIN_KEY: bin_name, INDEX_KEY: index, VALUE_KEY: count}

    if ctx:
        op_dict[CTX_KEY] = ctx

    return op_dict


def list_size(bin_name: str, ctx: Optional[list] = None):
    """Create a list size operation.

    Server returns the size of the list in the specified bin.

    Args:
        bin_name (str): The name of the bin containing the list.
        ctx (list): An optional list of nested CDT :class:`cdt_ctx <aerospike_helpers.cdt_ctx>` context operation
            objects.

    Returns:
        A dictionary usable in :meth:`~aerospike.Client.operate` and :meth:`~aerospike.Client.operate_ordered`. The
        format of the dictionary should be considered an internal detail, and subject to change.
    """
    op_dict = {OP_KEY: aerospike.OP_LIST_SIZE, BIN_KEY: bin_name}

    if ctx:
        op_dict[CTX_KEY] = ctx

    return op_dict


# Post 3.4.0 Operations. Require Server >= 3.16.0.1


def list_get_by_index(bin_name: str, index, return_type, ctx: Optional[list] = None):
    """Create a list get index operation.

    The list get operation gets the item at `index` and returns a value
    specified by `return_type`

    Args:
        bin_name (str): The name of the bin containing the list to fetch items from.
        index (int): The index of the item to be returned.
        return_type (int): Value specifying what should be returned from the operation.
            This should be one of the :ref:`list_return_types` values
        ctx (list): An optional list of nested CDT :class:`cdt_ctx <aerospike_helpers.cdt_ctx>` context operation
            objects.

    Returns:
        A dictionary usable in :meth:`~aerospike.Client.operate` and :meth:`~aerospike.Client.operate_ordered`. The
        format of the dictionary should be considered an internal detail, and subject to change.
    """
    op_dict = {
        OP_KEY: aerospike.OP_LIST_GET_BY_INDEX,
        BIN_KEY: bin_name,
        RETURN_TYPE_KEY: return_type,
        INDEX_KEY: index,
    }

    if ctx:
        op_dict[CTX_KEY] = ctx

    return op_dict


def list_get_by_index_range(bin_name: str, index, return_type, count=None, inverted=False, ctx: Optional[list] = None):
    """Create a list get index range operation.

    The list get by index range operation gets `count` items starting at `index` and returns a value
    specified by `return_type`

    Args:
        bin_name (str): The name of the bin containing the list to fetch items from.
        index (int): The index of the first item to be returned.
        return_type (int): Value specifying what should be returned from the operation.
            This should be one of the :ref:`list_return_types` values.
        count (int): The number of list items to be selected.
        inverted (bool): Optional bool specifying whether to invert the return type.
            If set to `True`, all items outside of the specified range will be returned.
            Default: `False`
        ctx (list): An optional list of nested CDT :class:`cdt_ctx <aerospike_helpers.cdt_ctx>` context operation
            objects.

    Returns:
        A dictionary usable in :meth:`~aerospike.Client.operate` and :meth:`~aerospike.Client.operate_ordered`. The
        format of the dictionary should be considered an internal detail, and subject to change.
    """
    op_dict = {
        OP_KEY: aerospike.OP_LIST_GET_BY_INDEX_RANGE,
        BIN_KEY: bin_name,
        RETURN_TYPE_KEY: return_type,
        INDEX_KEY: index,
        INVERTED_KEY: inverted,
    }

    if count is not None:
        op_dict[COUNT_KEY] = count

    if ctx:
        op_dict[CTX_KEY] = ctx

    return op_dict


def list_get_by_rank(bin_name: str, rank, return_type, ctx: Optional[list] = None):
    """Create a list get by rank operation.

    Server selects list item identified by `rank` and returns selected data
    specified by `return_type`.

    Args:
        bin_name (str): The name of the bin containing the list to fetch a value from.
        rank (int): The rank of the item to be fetched.
        return_type (int): Value specifying what should be returned from the operation.
            This should be one of the :ref:`list_return_types` values
        ctx (list): An optional list of nested CDT :class:`cdt_ctx <aerospike_helpers.cdt_ctx>` context operation
            objects.

    Returns:
        A dictionary usable in :meth:`~aerospike.Client.operate` and :meth:`~aerospike.Client.operate_ordered`. The
        format of the dictionary should be considered an internal detail, and subject to change.
    """
    op_dict = {OP_KEY: aerospike.OP_LIST_GET_BY_RANK, BIN_KEY: bin_name, RETURN_TYPE_KEY: return_type, RANK_KEY: rank}

    if ctx:
        op_dict[CTX_KEY] = ctx

    return op_dict


def list_get_by_rank_range(bin_name: str, rank, return_type, count=None, inverted=False, ctx: Optional[list] = None):
    """Create a list get by rank range operation.

    Server selects `count` items starting at the specified `rank` and returns selected data
    specified by `return_type`.

    Args:
        bin_name (str): The name of the bin containing the list to fetch items from.
        rank (int): The rank of the first items to be returned.
        return_type (int): Value specifying what should be returned from the operation.  This should be one of the
            :ref:`list_return_types` values
        count (int): A positive number indicating number of items to be returned.
        inverted (bool): Optional bool specifying whether to invert the return type.
            If set to `True`, all items outside of the specified rank range will be returned.
            Default: `False`
        ctx (list): An optional list of nested CDT :class:`cdt_ctx <aerospike_helpers.cdt_ctx>` context operation
            objects.

    Returns:
        A dictionary usable in :meth:`~aerospike.Client.operate` and :meth:`~aerospike.Client.operate_ordered`. The
        format of the dictionary should be considered an internal detail, and subject to change.
    """
    op_dict = {
        OP_KEY: aerospike.OP_LIST_GET_BY_RANK_RANGE,
        BIN_KEY: bin_name,
        RETURN_TYPE_KEY: return_type,
        RANK_KEY: rank,
        INVERTED_KEY: inverted,
    }

    if count is not None:
        op_dict[COUNT_KEY] = count

    if ctx:
        op_dict[CTX_KEY] = ctx

    return op_dict


def list_get_by_value(bin_name: str, value, return_type, inverted=False, ctx: Optional[list] = None):
    """Create a list get by value operation.

    Server selects list items with a value equal to `value` and returns selected data specified by
    `return_type`.

    Args:
        bin_name (str): The name of the bin containing the list to fetch items from.
        value: The server returns all items matching this value
        return_type (int): Value specifying what should be returned from the operation.  This should be one of the
            :ref:`list_return_types` values
        inverted (bool): Optional bool specifying whether to invert the return type.
            If set to `True`, all items not equal to `value` will be selected. Default: `False`
        ctx (list): An optional list of nested CDT :class:`cdt_ctx <aerospike_helpers.cdt_ctx>` context operation
            objects.
    Returns:
        A dictionary usable in :meth:`~aerospike.Client.operate` and :meth:`~aerospike.Client.operate_ordered`. The
        format of the dictionary should be considered an internal detail, and subject to change.
    """
    op_dict = {
        OP_KEY: aerospike.OP_LIST_GET_BY_VALUE,
        BIN_KEY: bin_name,
        RETURN_TYPE_KEY: return_type,
        VALUE_KEY: value,
        INVERTED_KEY: inverted,
    }

    if ctx:
        op_dict[CTX_KEY] = ctx

    return op_dict


def list_get_by_value_list(bin_name: str, value_list, return_type, inverted=False, ctx: Optional[list] = None):
    """Create a list get by value list operation.

    Server selects list items with a value contained in `value_list` and returns selected data
    specified by `return_type`.

    Args:
        bin_name (str): The name of the bin containing the list to fetch items from.
        value_list (list): Return items from the list matching an item in this list.
        return_type (int): Value specifying what should be returned from the operation.
            This should be one of the :ref:`list_return_types` values
        inverted (bool): Optional bool specifying whether to invert the return type.
            If set to `True`, all items not matching an entry in `value_list` will be selected.
            Default: `False`
        ctx (list): An optional list of nested CDT :class:`cdt_ctx <aerospike_helpers.cdt_ctx>` context operation
            objects.
    Returns:
        A dictionary usable in :meth:`~aerospike.Client.operate` and :meth:`~aerospike.Client.operate_ordered`. The
        format of the dictionary should be considered an internal detail, and subject to change.
    """
    op_dict = {
        OP_KEY: aerospike.OP_LIST_GET_BY_VALUE_LIST,
        BIN_KEY: bin_name,
        RETURN_TYPE_KEY: return_type,
        VALUE_LIST_KEY: value_list,
        INVERTED_KEY: inverted,
    }

    if ctx:
        op_dict[CTX_KEY] = ctx

    return op_dict


def list_get_by_value_range(
    bin_name: str,
    return_type,
    value_begin,
    value_end,
    inverted=False,
    ctx: Optional[list] = None
):
    """Create a list get by value list operation.

    Server selects list items with a value greater than or equal to `value_begin`
    and less than `value_end`. If `value_begin` is `None`, range is greater than or equal
    to the first element of the list. If `value_end` is `None` range extends to the end of the list.
    Server returns selected data specified by `return_type`.

    Args:
        bin_name (str): The name of the bin containing the list to fetch items from.
        return_type (int): Value specifying what should be returned from the operation.
            This should be one of the :ref:`list_return_types` values
        value_begin: The start of the value range.
        value_end: The end of the value range.
        inverted (bool): Optional bool specifying whether to invert the return type.
            If set to `True`, all items not included in the specified range will be returned.
            Default: `False`
        ctx (list): An optional list of nested CDT :class:`cdt_ctx <aerospike_helpers.cdt_ctx>` context operation
            objects.
    Returns:
        A dictionary usable in :meth:`~aerospike.Client.operate` and :meth:`~aerospike.Client.operate_ordered`. The
        format of the dictionary should be considered an internal detail, and subject to change.
    """
    op_dict = {
        OP_KEY: aerospike.OP_LIST_GET_BY_VALUE_RANGE,
        BIN_KEY: bin_name,
        RETURN_TYPE_KEY: return_type,
        INVERTED_KEY: inverted,
    }

    if value_begin is not None:
        op_dict[VALUE_BEGIN_KEY] = value_begin

    if value_end is not None:
        op_dict[VALUE_END_KEY] = value_end

    if ctx:
        op_dict[CTX_KEY] = ctx

    return op_dict


def list_remove_by_index(bin_name: str, index, return_type, ctx: Optional[list] = None):
    """Create a list remove by index operation.

    The list_remove_by_index operation removes the value of the item at `index` and returns a value
    specified by `return_type`

    Args:
        bin_name (str): The name of the bin containing the list to remove an item from.
        index (int): The index of the item to be removed.
        return_type (int): Value specifying what should be returned from the operation.
            This should be one of the :ref:`list_return_types` values
        ctx (list): An optional list of nested CDT :class:`cdt_ctx <aerospike_helpers.cdt_ctx>` context operation
            objects.

    Returns:
        A dictionary usable in :meth:`~aerospike.Client.operate` and :meth:`~aerospike.Client.operate_ordered`. The
        format of the dictionary should be considered an internal detail, and subject to change.
    """
    op_dict = {
        OP_KEY: aerospike.OP_LIST_REMOVE_BY_INDEX,
        BIN_KEY: bin_name,
        RETURN_TYPE_KEY: return_type,
        INDEX_KEY: index,
    }

    if ctx:
        op_dict[CTX_KEY] = ctx

    return op_dict


def list_remove_by_index_range(
    bin_name: str,
    index,
    return_type,
    count=None,
    inverted=False,
    ctx: Optional[list] = None
):
    """Create a list remove by index range operation.

    The list remove by index range operation removes `count` starting at `index` and returns a value
    specified by `return_type`.

    Args:
        bin_name (str): The name of the bin containing the list to remove items from.
        index (int): The index of the first item to be removed.
        return_type (int): Value specifying what should be returned from the operation.
            This should be one of the :ref:`list_return_types` values.
        count (int): The number of items to be removed
        inverted (bool): Optional bool specifying whether to invert the operation.
            If set to `True`, all items outside of the specified range will be removed.
            Default: `False`
        ctx (list): An optional list of nested CDT :class:`cdt_ctx <aerospike_helpers.cdt_ctx>` context operation
            objects.

    Returns:
        A dictionary usable in :meth:`~aerospike.Client.operate` and :meth:`~aerospike.Client.operate_ordered`. The
        format of the dictionary should be considered an internal detail, and subject to change.
    """
    op_dict = {
        OP_KEY: aerospike.OP_LIST_REMOVE_BY_INDEX_RANGE,
        BIN_KEY: bin_name,
        RETURN_TYPE_KEY: return_type,
        INDEX_KEY: index,
        INVERTED_KEY: inverted,
    }

    if count is not None:
        op_dict[COUNT_KEY] = count

    if ctx:
        op_dict[CTX_KEY] = ctx

    return op_dict


def list_remove_by_rank(bin_name: str, rank, return_type, ctx: Optional[list] = None):
    """Create a list remove by rank operation.

    Server removes a list item identified by `rank` and returns selected data
    specified by `return_type`.

    Args:
        bin_name (str): The name of the bin containing the list to fetch a value from.
        rank (int): The rank of the item to be removed.
        return_type (int): Value specifying what should be returned from the operation.
            This should be one of the :ref:`list_return_types` values
        ctx (list): An optional list of nested CDT :class:`cdt_ctx <aerospike_helpers.cdt_ctx>` context operation
            objects.

    Returns:
        A dictionary usable in :meth:`~aerospike.Client.operate` and :meth:`~aerospike.Client.operate_ordered`. The
        format of the dictionary should be considered an internal detail, and subject to change.
    """
    op_dict = {
        OP_KEY: aerospike.OP_LIST_REMOVE_BY_RANK,
        BIN_KEY: bin_name,
        RETURN_TYPE_KEY: return_type,
        RANK_KEY: rank,
    }

    if ctx:
        op_dict[CTX_KEY] = ctx

    return op_dict


def list_remove_by_rank_range(bin_name: str, rank, return_type, count=None, inverted=False, ctx: Optional[list] = None):
    """Create a list remove by rank range operation.

    Server removes `count` items starting at the specified `rank` and returns selected data
    specified by `return_type`.

    Args:
        bin_name (str): The name of the bin containing the list to fetch items from.
        rank (int): The rank of the first item to removed.
        return_type (int): Value specifying what should be returned from the operation.
            This should be one of the :ref:`list_return_types` values
        count (int): A positive number indicating number of items to be removed.
        inverted (bool): Optional bool specifying whether to invert the operation.
            If set to `True`, all items outside of the specified rank range will be removed.
            Default: `False`
        ctx (list): An optional list of nested CDT :class:`cdt_ctx <aerospike_helpers.cdt_ctx>` context operation
            objects.

    Returns:
        A dictionary usable in :meth:`~aerospike.Client.operate` and :meth:`~aerospike.Client.operate_ordered`. The
        format of the dictionary should be considered an internal detail, and subject to change.
    """
    op_dict = {
        OP_KEY: aerospike.OP_LIST_REMOVE_BY_RANK_RANGE,
        BIN_KEY: bin_name,
        RETURN_TYPE_KEY: return_type,
        RANK_KEY: rank,
        INVERTED_KEY: inverted,
    }

    if count is not None:
        op_dict[COUNT_KEY] = count

    if ctx:
        op_dict[CTX_KEY] = ctx

    return op_dict


def list_remove_by_value(bin_name: str, value, return_type, inverted=False, ctx: Optional[list] = None):
    """Create a list remove by value operation.

    Server removes list items with a value equal to `value` and returns selected data specified by
    `return_type`.

    Args:
        bin_name (str): The name of the bin containing the list to remove items from.
        value: The server removes all list items matching this value.
        return_type (int): Value specifying what should be returned from the operation.
            This should be one of the :ref:`list_return_types` values
        inverted (bool): Optional bool specifying whether to invert the operation.
            If set to `True`, all items not equal to `value` will be removed.
            Default: `False`
        ctx (list): An optional list of nested CDT :class:`cdt_ctx <aerospike_helpers.cdt_ctx>` context operation
            objects.
    Returns:
        A dictionary usable in :meth:`~aerospike.Client.operate` and :meth:`~aerospike.Client.operate_ordered`.The
        format of the dictionary should be considered an internal detail, and subject to change.
    """
    op_dict = {
        OP_KEY: aerospike.OP_LIST_REMOVE_BY_VALUE,
        BIN_KEY: bin_name,
        RETURN_TYPE_KEY: return_type,
        VALUE_KEY: value,
        INVERTED_KEY: inverted,
    }

    if ctx:
        op_dict[CTX_KEY] = ctx

    return op_dict


def list_remove_by_value_list(bin_name: str, value_list, return_type, inverted=False, ctx: Optional[list] = None):
    """Create a list remove by value list operation.

    Server removes list items with a value matching one contained in `value_list`
    and returns selected data specified by `return_type`.

    Args:
        bin_name (str): The name of the bin containing the list to remove items from.
        value_list (list): The server removes all list items matching one of these values.
        return_type (int): Value specifying what should be returned from the operation.
            This should be one of the :ref:`list_return_types` values
        inverted (bool): Optional bool specifying whether to invert the operation.
            If set to `True`, all items not equal to a value contained in `value_list` will be removed.
            Default: `False`
        ctx (list): An optional list of nested CDT :class:`cdt_ctx <aerospike_helpers.cdt_ctx>` context operation
            objects.
    Returns:
        A dictionary usable in :meth:`~aerospike.Client.operate` and :meth:`~aerospike.Client.operate_ordered`. The
        format of the dictionary should be considered an internal detail, and subject to change.
    """
    op_dict = {
        OP_KEY: aerospike.OP_LIST_REMOVE_BY_VALUE_LIST,
        BIN_KEY: bin_name,
        RETURN_TYPE_KEY: return_type,
        VALUE_LIST_KEY: value_list,
        INVERTED_KEY: inverted,
    }

    if ctx:
        op_dict[CTX_KEY] = ctx

    return op_dict


def list_remove_by_value_range(
    bin_name: str, return_type, value_begin=None, value_end=None, inverted=False, ctx: Optional[list] = None
):
    """Create a list remove by value range operation.

    Server removes list items with a value greater than or equal to `value_begin`
    and less than `value_end`.
    If `value_begin` is `None`, range is greater than or equal to the first element of the list.
    If `value_end` is `None` range extends to the end of the list.
    Server returns selected data specified by `return_type`.

    Args:
        bin_name (str): The name of the bin containing the list to fetch items from.
        return_type (int): Value specifying what should be returned from the operation.
            This should be one of the :ref:`list_return_types` values
        value_begin: The start of the value range.
        value_end: The end of the value range.
        inverted (bool): Optional bool specifying whether to invert the operation.
            If set to `True`, all items not included in the specified range will be removed.
            Default: `False`
        ctx (list): An optional list of nested CDT :class:`cdt_ctx <aerospike_helpers.cdt_ctx>` context operation
            objects.
    Returns:
        A dictionary usable in :meth:`~aerospike.Client.operate` and :meth:`~aerospike.Client.operate_ordered`.The
        format of the dictionary should be considered an internal detail, and subject to change.
    """
    op_dict = {
        OP_KEY: aerospike.OP_LIST_REMOVE_BY_VALUE_RANGE,
        BIN_KEY: bin_name,
        RETURN_TYPE_KEY: return_type,
        INVERTED_KEY: inverted,
    }

    if value_begin is not None:
        op_dict[VALUE_BEGIN_KEY] = value_begin

    if value_end is not None:
        op_dict[VALUE_END_KEY] = value_end

    if ctx:
        op_dict[CTX_KEY] = ctx

    return op_dict


def list_set_order(bin_name: str, list_order, ctx: Optional[list] = None):
    """Create a list set order operation.

    The list_set_order operation sets an order on a specified list bin.

    Args:
        bin_name (str): The name of the list bin.
        list_order: The ordering to apply to the list. Should be aerospike.LIST_ORDERED or
            aerospike.LIST_UNORDERED .
        ctx (list): An optional list of nested CDT :class:`cdt_ctx <aerospike_helpers.cdt_ctx>` context operation
            objects.
    Returns:
        A dictionary usable in :meth:`~aerospike.Client.operate` and :meth:`~aerospike.Client.operate_ordered`.The
        format of the dictionary should be considered an internal detail, and subject to change.
    """
    op_dict = {OP_KEY: aerospike.OP_LIST_SET_ORDER, BIN_KEY: bin_name, LIST_ORDER_KEY: list_order}

    if ctx:
        op_dict[CTX_KEY] = ctx

    return op_dict


def list_sort(bin_name: str, sort_flags: int = 0, ctx: Optional[list] = None):
    """Create a list sort operation

    The list sort operation will sort the specified list bin.

    Args:
        bin_name (str): The name of the bin to sort.
        sort_flags (int): :ref:`aerospike_list_sort_flag` modifying the sorting behavior
            (default ``aerospike.DEFAULT_LIST_SORT``).
        ctx (list): An optional list of nested CDT :class:`cdt_ctx <aerospike_helpers.cdt_ctx>` context operation
            objects.
    Returns:
        A dictionary usable in :meth:`~aerospike.Client.operate` and :meth:`~aerospike.Client.operate_ordered`.The
        format of the dictionary should be considered an internal detail, and subject to change.
    """
    op_dict = {OP_KEY: aerospike.OP_LIST_SORT, BIN_KEY: bin_name, SORT_FLAGS_KEY: sort_flags}

    if ctx:
        op_dict[CTX_KEY] = ctx

    return op_dict


def list_get_by_value_rank_range_relative(
    bin_name: str, value, offset, return_type, count=None, inverted=False, ctx: Optional[list] = None
):
    """Create a list get by value rank range relative operation

    Create list get by value relative to rank range operation.
    Server selects list items nearest to value and greater by relative rank.
    Server returns selected data specified by return_type.

    Args:
        bin_name (str): The name of the bin containing the list.
        value (str): The value of the item in the list for which to search
        offset (int): Begin returning items with rank == rank(found_item) + offset
        return_type (int): Value specifying what should be returned from the operation.
            This should be one of the :ref:`list_return_types` values
        count (int): If specified, the number of items to return. If None,
            all items until end of list are returned.
        inverted (bool): If True, the operation is inverted, and items outside
            of the specified range are returned.
        ctx (list): An optional list of nested CDT :class:`cdt_ctx <aerospike_helpers.cdt_ctx>` context operation
            objects.

    Returns:
        A dictionary usable in :meth:`~aerospike.Client.operate` and :meth:`~aerospike.Client.operate_ordered`.The
        format of the dictionary should be considered an internal detail, and subject to change.

    Note:
        This operation requires server version 4.3.0 or greater.

    Examples:
        These examples show what would be returned for specific arguments
        when dealing with an ordered list: ``[0,4,5,9,11,15]``

        ::

            (value, offset, count) = [selected items]
            # Found 5
            # Rank 0 = 5
            # No upper limit, so get all elements >= 5
            (5, 0, None) = [5,9,11,15]
            # Only get 2 elements
            (5, 0, 2) = [5, 9]

            # Relative rank -1 = 4
            # since 4 is just below 5 in the list
            # Then get all elements >= 4
            (5, -1, None) = [4, 5, 9, 11, 15]
            # Only get 3 elements
            (5, -1, 3) = [4, 5, 9]

            # 3 not in list
            # But next greater value is 4, so set 4's rank to 0
            # Then rank 0 = 4, rank 1 = 5, rank 2 = 9, rank 3 = 11
            (3,3, None) = [11, 15]
            # Rank 0 = 4, Rank -1 = 0
            # We will only get down to the lowest rank item
            (3,-3, None) = [0, 4,5,9,11,15]
            (3, 0, None) = [4,5,9,11,15]

    """
    op_dict = {
        OP_KEY: aerospike.OP_LIST_GET_BY_VALUE_RANK_RANGE_REL,
        BIN_KEY: bin_name,
        VALUE_KEY: value,
        RANK_KEY: offset,
        RETURN_TYPE_KEY: return_type,
        INVERTED_KEY: inverted,
    }

    if count is not None:
        op_dict[COUNT_KEY] = count

    if ctx:
        op_dict[CTX_KEY] = ctx

    return op_dict


def list_remove_by_value_rank_range_relative(
    bin_name: str, value, offset, return_type, count=None, inverted=False, ctx: Optional[list] = None
):
    """Create a list get by value rank range relative operation

    Create list remove by value relative to rank range operation.
    Server removes and returns list items nearest to value and greater by relative rank.
    Server returns selected data specified by return_type.

    Args:
        bin_name (str): The name of the bin containing the list.
        value (str): The value of the item in the list for which to search
        offset (int): Begin removing and returning items with rank == rank(found_item) + offset
        return_type (int): Value specifying what should be returned from the operation.
            This should be one of the :ref:`list_return_types` values
        count (int): If specified, the number of items to remove and return. If None,
            all items until end of list are returned.
        inverted (bool): If True, the operation is inverted, and items outside of the specified
            range are removed and returned.
        ctx (list): An optional list of nested CDT :class:`cdt_ctx <aerospike_helpers.cdt_ctx>` context operation
            objects.

    Returns:
        A dictionary usable in :meth:`~aerospike.Client.operate` and :meth:`~aerospike.Client.operate_ordered`.The
        format of the dictionary should be considered an internal detail, and subject to change.

    Note:
        This operation requires server version 4.3.0 or greater.

    These examples show what would be removed and returned for specific arguments
    when dealing with an ordered list: ``[0,4,5,9,11,15]``

    ::

        (value, offset, count) = [selected items]
        (5,0,None) = [5,9,11,15]
        (5,0,2) = [5, 9]
        (5,-1, None) = [4,5,9,11,15]
        (5, -1, 3) = [4,5,9]
        (3,3, None) = [11,15]
        # We can only go down to lowest rank item
        (3,-3, None) = [0,4,5,9,11,15]
        (3, 0, None) = [4,5,9,11,15]

    """
    op_dict = {
        OP_KEY: aerospike.OP_LIST_REMOVE_BY_VALUE_RANK_RANGE_REL,
        BIN_KEY: bin_name,
        VALUE_KEY: value,
        RANK_KEY: offset,
        RETURN_TYPE_KEY: return_type,
        INVERTED_KEY: inverted,
    }
    if count is not None:
        op_dict[COUNT_KEY] = count

    if ctx:
        op_dict[CTX_KEY] = ctx

    return op_dict



class AerospikeError(Exception):
    code: int
    msg: str
    file: str
    line: int
    in_doubt: bool

class ClientError(AerospikeError):
    pass

class InvalidHostError(ClientError):
    pass

class ParamError(ClientError):
    pass

class ServerError(AerospikeError):
    pass

class InvalidRequest(ServerError):
    pass

class OpNotApplicable(ServerError):
    pass

class FilteredOut(ServerError):
    pass

class ServerFull(ServerError):
    pass

class AlwaysForbidden(ServerError):
    pass

class UnsupportedFeature(ServerError):
    pass

class DeviceOverload(ServerError):
    pass

class NamespaceNotFound(ServerError):
    pass

class ForbiddenError(ServerError):
    pass

class ElementExistsError(ServerError):
    pass

class ElementNotFoundError(ServerError):
    pass

class RecordError(ServerError):
    key: tuple
    bin: str

class RecordKeyMismatch(RecordError):
    pass

class RecordNotFound(RecordError):
    pass

class RecordGenerationError(RecordError):
    pass

class RecordExistsError(RecordError):
    pass

class RecordBusy(RecordError):
    pass

class RecordTooBig(RecordError):
    pass

class BinNameError(RecordError):
    pass

class BinIncompatibleType(RecordError):
    pass

class IndexError(ServerError):
    pass

class IndexNotFound(IndexError):
    pass

class IndexFoundError(IndexError):
    pass

class IndexOOM(IndexError):
    pass

class IndexNotReadable(IndexError):
    pass

class IndexNameMaxLen(IndexError):
    pass

class IndexNameMaxCount(IndexError):
    pass

class QueryError(AerospikeError):
    pass

class QueryQueueFull(QueryError):
    pass

class QueryTimeout(QueryError):
    pass

class ClusterError(AerospikeError):
    pass

class ClusterChangeError(ClusterError):
    pass

class AdminError(ServerError):
    pass

class ExpiredPassword(AdminError):
	pass

class ForbiddenPassword(AdminError):
	pass

class IllegalState(AdminError):
	pass

class InvalidCommand(AdminError):
	pass

class InvalidCredential(AdminError):
	pass

class InvalidField(AdminError):
	pass

class InvalidPassword(AdminError):
	pass

class InvalidPrivilege(AdminError):
	pass

class InvalidRole(AdminError):
	pass

class InvalidUser(AdminError):
	pass

class NotAuthenticated(AdminError):
	pass

class RoleExistsError(AdminError):
	pass

class RoleViolation(AdminError):
	pass

class SecurityNotEnabled(AdminError):
	pass

class SecurityNotSupported(AdminError):
	pass

class SecuritySchemeNotSupported(AdminError):
	pass

class UserExistsError(AdminError):
	pass

class UDFError(ServerError):
    module: str
    func: str

class UDFNotFound(UDFError):
    pass

class LuaFileNotFound(UDFError):
    pass



##########################################################################
# Copyright 2013-2021 Aerospike, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##########################################################################
"""
Bitwise operator expressions provide support for bitwise operators like `&` and `>>` in Aerospike expressions.

"""

# from __future__ import annotations
from typing import Union

from aerospike_helpers.expressions.resources import _GenericExpr
from aerospike_helpers.expressions.resources import _BaseExpr
from aerospike_helpers.expressions.resources import _ExprOp

##############################
# Bitwise Operator Expressions
##############################

TypeInteger = Union[_BaseExpr, int]
TypeBool = Union[_BaseExpr, bool]


class IntAnd(_BaseExpr):
    """Create integer "and" (&) operator expression that is applied to two or more integers.
    All arguments must resolve to integers.

    Requires server version 5.6.0+.
    """

    _op = _ExprOp.INT_AND

    def __init__(self, *exprs: "TypeInteger"):
        """Args:
            `*exprs` (TypeInteger): A variable amount of integer expressions or values to be bitwise ANDed.

        :return: (integer value)

        Example::

            # for int bin "a", a & 0xff == 0x11
            expr = exp.Eq(exp.IntAnd(exp.IntBin("a"), 0xff), 0x11).compile()
        """
        self._children = exprs + (_GenericExpr(_ExprOp._AS_EXP_CODE_END_OF_VA_ARGS, 0, {}),)


class IntOr(_BaseExpr):
    """Create integer "or" (|) operator expression that is applied to two or more integers.
    All arguments must resolve to integers.

    Requires server version 5.6.0+.
    """

    _op = _ExprOp.INT_OR

    def __init__(self, *exprs: "TypeInteger"):
        """Args:
            `*exprs` (TypeInteger): A variable amount of integer expressions or values to be bitwise ORed.

        :return: (integer value)

        Example::

            # for int bin "a", a | 0x10 not == 0
            expr = exp.NE(exp.IntOr(IntBin("a"), 0x10), 0).compile()
        """
        self._children = exprs + (_GenericExpr(_ExprOp._AS_EXP_CODE_END_OF_VA_ARGS, 0, {}),)


class IntXOr(_BaseExpr):
    """Create integer "xor" (^) operator that is applied to two or more integers.
    All arguments must resolve to integers.

    Requires server version 5.6.0+.
    """

    _op = _ExprOp.INT_XOR

    def __init__(self, *exprs: "TypeInteger"):
        """Args:
            `*exprs` (TypeInteger): A variable amount of integer expressions or values to be bitwise XORed.

        :return: (integer value)

        Example::

            # for int bin "a", "b", a ^ b == 16
            expr = exp.Eq(exp.IntXOr(exp.IntBin("a"), exp.IntBin("b")), 16).compile()
        """
        self._children = exprs + (_GenericExpr(_ExprOp._AS_EXP_CODE_END_OF_VA_ARGS, 0, {}),)


class IntNot(_BaseExpr):
    """Create integer "not" (~) operator.

    Requires server version 5.6.0+.
    """

    _op = _ExprOp.INT_NOT

    def __init__(self, expr: "TypeInteger"):
        """Args:
            `expr` (TypeInteger): An integer value or expression to be bitwise negated.

        :return: (integer value)

        Example::

            # for int bin "a", ~ a == 7
            expr = exp.Eq(exp.IntNot(exp.IntBin("a")), 7).compile()
        """
        self._children = (expr,)


class IntLeftShift(_BaseExpr):
    """Create integer "left shift" (<<) operator.

    Requires server version 5.6.0+.
    """

    _op = _ExprOp.INT_LSHIFT

    def __init__(self, value: "TypeInteger", shift: "TypeInteger"):
        """Args:
            `value` (TypeInteger): An integer value or expression to be left shifted.
            `shift` (TypeInteger): An integer value or expression for number of bits to left shift `value` by.

        :return: (integer value)

        Example::

            # for int bin "a", a << 8 > 0xff
            expr = exp.GT(exp.IntLeftShift(exp.IntBin("a"), 8), 0xff).compile()
        """
        self._children = (value, shift)


class IntRightShift(_BaseExpr):
    """Create integer "logical right shift" (>>>) operator.

    Requires server version 5.6.0+.
    """

    _op = _ExprOp.INT_RSHIFT

    def __init__(self, value: "TypeInteger", shift: "TypeInteger"):
        """Args:
            `value` (TypeInteger): An integer value or expression to be right shifted.
            `shift` (TypeInteger): An integer value or expression for number of bits to right shift `value` by.

        :return: (integer value)

        Example::

            # for int bin "a", a >>> 8 > 0xff
            expr = exp.GT(exp.IntRightShift(exp.IntBin("a"), 8), 0xff).compile()
        """
        self._children = (value, shift)


class IntArithmeticRightShift(_BaseExpr):
    """Create integer "arithmetic right shift" (>>) operator.

    Requires server version 5.6.0+.
    """

    _op = _ExprOp.INT_ARSHIFT

    def __init__(self, value: "TypeInteger", shift: "TypeInteger"):
        """Args:
            `value` (TypeInteger): An integer value or expression to be right shifted.
            `shift` (TypeInteger): An integer value or expression for number of bits to right shift `value` by.

        :return: (integer value)

        Example::

            # for int bin "a", a >> 8 > 0xff
            expr = exp.GT(exp.IntArithmeticRightShift(exp.IntBin("a"), 8), 0xff).compile()
        """
        self._children = (value, shift)


class IntCount(_BaseExpr):
    """Create expression that returns count of integer bits that are set to 1.

    Requires server version 5.6.0+.
    """

    _op = _ExprOp.INT_COUNT

    def __init__(self, value: "TypeInteger"):
        """Args:
            `value` (TypeInteger): An integer value or expression to have bits counted.

        :return: (integer value)

        Example::

            # for int bin "a", count(a) == 4
            expr = exp.Eq(exp.IntCount(exp.IntBin("a")), 4).compile()
        """
        self._children = (value,)


class IntLeftScan(_BaseExpr):
    """Create expression that scans integer bits from left (most significant bit) to
    right (least significant bit), looking for a search bit value. When the
    search value is found, the index of that bit (where the most significant bit is
    index 0) is returned. If "search" is true, the scan will search for the bit
    value 1. If "search" is false it will search for bit value 0.

    Requires server version 5.6.0+.
    """

    _op = _ExprOp.INT_LSCAN

    def __init__(self, value: "TypeInteger", search: "TypeBool"):
        """Args:
            `value` (TypeInteger): An integer value or expression to be scanned.
            `search` (TypeBool): A bool expression or value to scan for.

        :return: (integer value)

        Example::

            # for int bin "a", lscan(a, True) == 4
            expr = exp.GT(exp.IntLeftScan(exp.IntBin("a"), True), 4).compile()
        """
        self._children = (value, search)


class IntRightScan(_BaseExpr):
    """
    Create expression that scans integer bits from right (least significant bit) to
    left (most significant bit), looking for a search bit value. When the
    search value is found, the index of that bit (where the most significant bit is
    index 0) is returned. If "search" is true, the scan will search for the bit
    value 1. If "search" is false it will search for bit value 0.

    Requires server version 5.6.0+.
    """

    _op = _ExprOp.INT_RSCAN

    def __init__(self, value: "TypeInteger", search: "TypeBool"):
        """Args:
            `value` (TypeInteger): An integer value or expression to be scanned.
            `search` (TypeBool): A bool expression or value to scan for.

        :return: (integer value)

        Example::

            # for int bin "a", rscan(a, True) == 4
            expr = exp.GT(exp.IntRightScan(exp.IntBin("a"), True), 4).compile()
        """
        self._children = (value, search)
